/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.animation;

import java.awt.DisplayMode;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.util.Vector;

/**
 * 
 * This class does all the animation support.
 * 
 * The idea is simply to replace floats by AnimatedFloats, you use set() and get()
 * and that's basically it. I don't think you need to look deep into this code.
 * 
 * @author dragice
 *
 */
public class AnimatedFloat { //cobertura:ignore
	
	public static final double DEFAULT_QUADRATIC_INCREMENT = 0.2f;
	public static final double DEFAULT_LINEAR_INCREMENT = 0.001f;
	
	private double value;
	
	private double destValue;
	private double currentLinearIncrement = 0;
	private double currentQuadraticIncrement = 0;
	
	private double quadraticIncrement;
	private double linearIncrement;
	private double quadraticDecrement;
	private double linearDecrement;
	private boolean valueChanged = false;

	private static Vector<AnimatedFloat> allFloats = new Vector<AnimatedFloat>();
	
	static boolean animationEnabled = true;
	static boolean framerateCorrectionEnabled = true;
	static private long lasttime = 0;
	
	public AnimatedFloat(double initialValue) {
		this(initialValue, DEFAULT_QUADRATIC_INCREMENT, DEFAULT_LINEAR_INCREMENT);
	}

	public AnimatedFloat(double initialValue, double quadraticIncrement, double linearIncrement) {
		this(initialValue, quadraticIncrement, linearIncrement, quadraticIncrement, linearIncrement);
	}
	
	public AnimatedFloat(double initialValue, double quadraticIncrement, double linearIncrement, double quadraticDecrement, double linearDecrement) {
		this.value = initialValue;
		this.destValue = value;
		this.quadraticIncrement = quadraticIncrement;
		this.linearIncrement = linearIncrement;
		this.quadraticDecrement = quadraticDecrement;
		this.linearDecrement = linearDecrement;
		allFloats.add(this);
	}
	
	public void dispose() {
		allFloats.remove(this);
	}
	
	public void setIncrement(double quadraticIncrement, double linearIncrement) {
		setIncrement(quadraticIncrement, linearIncrement, quadraticIncrement, linearIncrement);
	}
	
	public double getQuadraticIncrement() {
		return quadraticIncrement;
	}
	
	public double getLinearIncrement() {
		return linearIncrement;
	}

	public void setIncrement(double quadraticIncrement, double linearIncrement, double quadraticDecrement, double linearDecrement) {
		this.quadraticIncrement = quadraticIncrement;
		this.linearIncrement = linearIncrement;
		this.quadraticDecrement = quadraticDecrement;
		this.linearDecrement = linearDecrement;
		if (value != destValue) {
			boolean forward = (destValue > value);
			currentLinearIncrement = (destValue - value) * (forward ? linearIncrement : linearDecrement);
			currentQuadraticIncrement = forward ? quadraticIncrement : quadraticDecrement;
		}
	}
	
	@Override
	public AnimatedFloat clone() {
		AnimatedFloat clone = new AnimatedFloat(0, quadraticIncrement, linearIncrement, quadraticDecrement, linearDecrement);
		clone.value = value;
		clone.destValue = value;
		clone.currentLinearIncrement = currentLinearIncrement;
		clone.currentQuadraticIncrement = currentQuadraticIncrement;
		clone.valueChanged = valueChanged;
		return clone;
	}
	
	public double get() {
		return value;
	}
	
	public double directGet() {
		return destValue;
	}
	
	public int getInt() {
		return (int)Math.round(value);
	}
	
	public void set(double v) {
		if (!animationEnabled)
			directSet(v);
		else {
			if (destValue == v) return;
			destValue = v;
			boolean forward = (destValue > value);
			currentLinearIncrement = (destValue - value) * (forward ? linearIncrement : linearDecrement);
			currentQuadraticIncrement = forward ? quadraticIncrement : quadraticDecrement;
			AnimationTimer.maybeAnimate();
		}
	}
	
	public void directSet(double v) {
		destValue = v;
		value = v;
		currentLinearIncrement = 0;
		currentQuadraticIncrement = 0;
		valueChanged = true;
	}
	
	public void add(double v) {
		set(destValue + v);
	}
	
	public void directAdd(double v) {
		directSet(destValue + v);
	}

	public void mul(double v) {
		set(destValue * v);
	}
	
	public void directMul(double v) {
		directSet(destValue * v);
	}

	public boolean changed() {
		return valueChanged;
	}
	
	public boolean animate() {
		return animate(1);
	}
	
	private boolean animate(double timeStretch) {
		if (destValue == value) {
			currentLinearIncrement = 0;
			currentQuadraticIncrement = 0;
			valueChanged = false;
			return valueChanged;
		}
		
		double newValue = value + (destValue - value) * currentQuadraticIncrement * timeStretch + (timeStretch * currentLinearIncrement);
		boolean overshoot = (newValue - destValue) * (value - destValue) < 0;
		if (overshoot)
			value = destValue;
		else
			value = newValue;
		valueChanged = true;
		return valueChanged;
	}
	
	boolean _needsAnimation() {
		return value != destValue;
	}
	
	public boolean endAnimation() {
		if (destValue == value)
			return false;
		
		value = destValue;
		currentLinearIncrement = 0;
		currentQuadraticIncrement = 0;
		return true;
	}
	
	public static void setAnimationEnabled(boolean enabled) {
		animationEnabled = enabled;
	}
	
	public static boolean animateAll() {
		double timeStretch = getTimeStretch();
		int N = allFloats.size();
		AnimatedFloat f;
		boolean needsMoreAnimation = false;
		for (int i=0; i<N; i++) {
			f = allFloats.elementAt(i);
			needsMoreAnimation |= f.animate(timeStretch);
		}
		if (!needsMoreAnimation)
			lasttime = 0;
		return needsMoreAnimation;
	}
	
	public static boolean needsAnimation() {
		int N = allFloats.size();
		for (int i=0; i<N; i++)
			if (allFloats.elementAt(i)._needsAnimation())
				return true;
		return false;
	}
	
	public static void removeAllAnimations(){
	    allFloats.clear();
	}

	// This is the (little bit complicated) code for framerate correction
	public static double getTimeStretch() {
		double timeStretch = 1;
		if (framerateCorrectionEnabled) { 
			long newtime = System.nanoTime();
			if (lasttime == 0) {
				lasttime = newtime;
				return 1;
			}
			double elapsed_seconds = (newtime-lasttime)/1000000000.0;
			timeStretch = elapsed_seconds / (AnimationTimer.ANIMATION_RATE / 1000.0);
			lasttime = newtime;
		}
		return timeStretch;
	}
	
	public double estimateAnimationTime(float deltaValue) {
		if (deltaValue == 0)
			return 0;
		boolean forward = deltaValue > 0;
		double currentLinearIncrement_ = deltaValue * (forward ? linearIncrement : linearDecrement);
		double currentQuadraticIncrement_ = forward ? quadraticIncrement : quadraticDecrement;

		if (currentLinearIncrement_ == 0)
			return Double.MAX_VALUE;
		if (currentQuadraticIncrement_ == 0)
			return Double.MIN_VALUE;
		double ratio = currentLinearIncrement_ / currentQuadraticIncrement_;
		double time = - Math.log(ratio / (deltaValue + ratio)) / currentQuadraticIncrement_;
		return time * AnimationTimer.ANIMATION_RATE / 1000.0;
	}
}
