/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.diffanimation;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.Toolkit;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.geom.AffineTransform;
import java.io.File;
import java.util.ArrayList;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import fr.aviz.animation.AnimationListener;
import fr.aviz.animation.AnimationTimer;
import fr.aviz.diff.DiffUtils;
import fr.aviz.diffanimation.DocumentAnimator.PaintMode;
import fr.aviz.diffanimation.gui.GUIUtils;
import fr.aviz.diffanimation.gui.JAnimationSettings;
import fr.aviz.diffanimation.gui.JRevisionSelector;
import fr.aviz.diffanimation.overview.DiffOverview;
import fr.aviz.diffanimation.timeline.CompoundDiffTimeline;
import fr.aviz.utils.DocumentFactory;
import fr.aviz.utils.ExecTime;
import fr.aviz.utils.WordAdvanceCache;


public class DiffAnimator extends JComponent implements AnimationListener, ComponentListener, GUIUtils.AdvancedKeyListener {

	/* Default example */
	public static final String defaultFile = ""; //"ivtk";//"ivtk";//"CHI_(conference)";
	
	public static final boolean WORD_BOUNDARIES_DEBUG = false;
	public static final boolean BLOCK_BOUNDARIES_DEBUG = false;
	public static final boolean PAR_BOUNDARIES_DEBUG = false;
	public static final boolean HIGHLIGHTS_BOUNDARIES_DEBUG = false;	
	
	public static final boolean WORD_DISPLACEMENT_DEBUG = false;	
	public static Color BG_COLOR = new Color(0.93f, 0.93f, 0.95f);
	public static Color DOC_COLOR = new Color(1f, 1f, 1f);

	//public static Version DEFAULT_START = Version.V1;
	//public static Version DEFAULT_END = Version.V2;

	/**
	 * An animation between two text files.
	 * These objects are queued in animationQueue to allow chained animations.
	 */
	private static class Animation {	
		private final DocumentAnimator document;
		//private boolean forward
		private boolean animating = false;
		private float t0;
		private float t1;
		
		public Animation(String file1, String file2, float t0, float t1, Settings settings) {
			this.t0 = t0;
			this.t1 = t1;
			this.document = DocumentFactory.createDocument(file1, file2);
		//	this.forward = forward;
		}
		
		public void initialize(boolean turbo) {
			document.setAnimationState(t0);
			document.setTurbo(turbo);
		}
		
		public void start() {
			if (!animating) {
				animating = true;
				document.startAnimation(t0, t1);
			}
		}
		
		public void stop() {
			if (animating) {
				document.endAnimation();
				animating = false;
			}
		}
		
		public void setTurbo(boolean turbo) {
			document.setTurbo(turbo);
			document.updateSpeed();
		}
		
		// gives the opportunity to dynamically reverse the animation's direction of motion
		public void setDestination(float t1) {
			this.t1 = t1;
			if (animating) {
				document.changeAnimation(t1);
			}
		}
	}
	
	ArrayList<Animation> animationQueue = new ArrayList<Animation>();
	Animation currentAnimation = null;
	
//	private DocumentAnimator document;
	private DiffOverview overview;
	private CompoundDiffTimeline timeline;
	private static int frameWidth = 700;
	public Settings settings;
	
	float translatey = 0;
	float scale = 1f;
	
	String[] allFiles = null;
	String[] allFilesShortestPath = null;

	final JFrame win; // used for changing frame title
	
	/**
	 * The important thing here is the call to AnimationTimer.addListener().
	 */
	public DiffAnimator(JFrame win) {
		super();
		this.win = win;
		setPreferredSize(new Dimension(frameWidth,800));
		setFocusable(true);
		AnimationTimer.addListener(this);
		addComponentListener(this);
		settings = new Settings();
		DocumentFactory.setSettings(settings);
		GUIUtils.addAdvancedKeyListener(null, this, true);
	}
	
	public void setSpeedNormalizer(SpeedNormalizer normalizer) {
		settings.speedNormalizer = normalizer;
	}
	
	public SpeedNormalizer getSpeedNormalizer() {
		return settings.speedNormalizer;
	}
	
	public void setOverview(DiffOverview overview) {
		this.overview = overview;
	}
	
	public void setTimeline(CompoundDiffTimeline timeline) {
		this.timeline = timeline;
	}

	/**
	 * Called when a new set of documents has been selected
	 */
	public void newFiles(String[] files) {
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		emptyAnimationQueue();
		allFiles = files;
		allFilesShortestPath = null;
		// Empty data from caches
		DocumentFactory.clearCache();
		WordAdvanceCache.clearCache();
		System.gc();
		// Create document transitions in advance (also used for computing max height)
		DocumentFactory.createDocuments(allFiles); // create and cache documents to speed up (optional)
		// Init overview
		overview.fitToFiles(files);
		overview.updateOpacity();
		timeline.setMaxDocumentHeight(overview.getMaxDocumentHeight());
		setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
		requestFocus();
	}
	
	public void computeShortestPath() {
		allFilesShortestPath = DiffUtils.getShortestDiffPath(allFiles, settings.diffMethod);
		DocumentFactory.createDocuments(allFilesShortestPath);
	}
	
	/**
	 * Called when a diff method has been changed
	 */
	public void rebuildAllDocuments() {
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		DocumentAnimator old_doc = getDocument();
		newFiles(allFiles);
		if (old_doc != null) {
			startAnimation(old_doc.getFilename1(), old_doc.getFilename2(), 0, 0);
			emptyAnimationQueue();
			setAnimationState(old_doc.getAnimationState(), false);
			repaint();
		}
		timeline.layoutMarks();
		setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	}
	
	public String[] getCurrentFiles() {
		return allFiles;
	}
	
	public String[] getCurrentFilesShortestPath() {
		return allFilesShortestPath;
	}

	private Animation getLastAnimation() {
		if (animationQueue.size() == 0)
			return currentAnimation;
		else
			return animationQueue.get(animationQueue.size() - 1);
	}
	
	public String getLastFilename() {
		Animation lastanim = getLastAnimation();
		if (lastanim != null) {
			if (lastanim.t1 > 0.5)
				return lastanim.document.getFilename2();
			else
				return lastanim.document.getFilename1();
		}
		return null;
	}
	
	public String getCurrentFilename() {
		if (currentAnimation == null)
			return "";
		if (currentAnimation.document.getAnimationState() < 0.5)
			return currentAnimation.document.getFilename1();
		else
			return currentAnimation.document.getFilename2();
	}
	
	public int getIndex(String filename) {
		for (int i=0; i<allFiles.length; i++)
			if (allFiles[i].equals(filename))
				return i;
		return -1;
	}
	
	/**
	 * Queues an animation from the last file in the queue to file2.
	 */
	public void queueAnimation(String file2) {
		String file1 = getLastFilename();
		if (file1 == null)
			file1 = file2;
		queueAnimation(file1, file2);
	}
	
	/**
	 * Queues an animation from file1 to file2.
	 */
	public void queueAnimation(String file1, String file2) {
		if (getIndex(file2) >= getIndex(file1))
			queueAnimation(file1, file2, 0, 1);
		else
			queueAnimation(file2, file1, 1, 0);
	}
	
	/**
	 * Queues an animation. If no animation is playing, will play the animation right away.
	 * 
	 * IMPORTANT: file1's revision should be lower than file2's.
	 */
	public void queueAnimation(String file1, String file2, float t0, float t1) {
		// If an animation between the same files is already playing and it is the last one, just update it
		if (animationQueue.size() == 1 && currentAnimation != null && currentAnimation.document.equals(file1, file2) && currentAnimation.animating) {
			currentAnimation.setTurbo(false);
			currentAnimation.setDestination(t1);
		}
		// Otherwise, add a new animation
		else {
			animationQueue.add(new Animation(file1, file2, t0, t1, settings));
			if (animationQueue.size() == 1)
				// this is the first animation in the queue
				animateNext();
			else {
				Animation previousAnim = animationQueue.get(animationQueue.size() - 2);
				previousAnim.setTurbo(true);
				if (previousAnim == currentAnimation)
					previousAnim.document.updateSpeed();
			}
		}
	}
	
	/**
	 * Plays an animation right away.
	 * 
	 * @param file1
	 * @param file2
	 */
	public void startAnimation(String file1, String file2) {
		if (getIndex(file2) >= getIndex(file1))
			startAnimation(file1, file2, 0, 1);
		else
			startAnimation(file2, file1, 1, 0);
	}
	
	/**
	 * Plays an animation right away, no matter what's currently playing.
	 * 
	 * IMPORTANT: file1's revision should be lower than file2's.
	 */
	public void startAnimation(String file1, String file2, float t0, float t1) {
		// If an animation between the same files is already playing, just update it
		if (currentAnimation != null && currentAnimation.document.equals(file1, file2) && currentAnimation.animating) {
			currentAnimation.setTurbo(false);
			currentAnimation.setDestination(t1);
		}
		// Otherwise, stop and clear everything, then add a new animation.
		else {
			emptyAnimationQueue();
			queueAnimation(file1, file2, t0, t1);
			animateNext();
		}
	}
	
	public void boostCurrentAnimation(boolean turbo) {
		if (currentAnimation != null) {
			currentAnimation.setTurbo(turbo);
			if (currentAnimation.animating)
				currentAnimation.document.updateSpeed();
		}
	}
	
/*	public void showRevision(String file) {
		emptyAnimationQueue();
		currentAnimation = new Animation(file, file, forward, settings);
		animationQueue.add(currentAnimation);
		
	}
	*/
	public void clear() {
		emptyAnimationQueue();
		currentAnimation = null;
		scale = 1;
		translatey = 0;
		//overview.setT
		overview.moveThumbManually(0);
		overview.directUpdateOpacity();
		repaint();
		overview.repaint();
		timeline.clear();
	}

	/**
	 * Stops the currently playing animation (if any) and empties the animation queue.
	 */
	public void emptyAnimationQueue() {
		
		if (currentAnimation != null)
			currentAnimation.stop();
		animationQueue.clear();

	}
	
	private void animateNext() {
		
		if (animationQueue.isEmpty())
			return;
		
		if (animationQueue.get(0) == currentAnimation) {
			if (currentAnimation.animating)
				return;
			else {
				animationQueue.remove(0);
			}
		}
		
		///////////////////////
		// This is where the v1/v2 document is actually changing.
		///////////////////////
		currentAnimation = animationQueue.get(0);
		currentAnimation.initialize(animationQueue.size() > 1);
		repaint();
		// it's important the overview is notified first because it needs to stabilize the document
		// before visible motion is computed.
		overview.documentChanged(currentAnimation.document);
		settings.speedNormalizer.documentChanged(currentAnimation.document);
		timeline.updateCursorPosition();
		timeline.repaint();
		//overview.stabilize();
		currentAnimation.start();
		animateEvent();
		//overview.stabilize();
	}
	
	/**
	 * This will tell the component to repaint itself each time an animated float
	 * changes. See the call to AnimationTimer.addListener in the constructor.
	 */
	public void animateEvent() {
		// Repaint
		repaint();
		overview.repaint();
		timeline.updateCursorPosition();

		// Start a new animation?
		if (animationQueue.size() > 0) {
			final Animation a = animationQueue.get(0);
			if (!a.document.animating()) {
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						animationQueue.remove(a);
						animateNext();
					}
				});
			}
		}
		
		updateWindowTitle();
	}
	
	
	public void updateWindowTitle() {
		setWindowTitle(getCurrentFilename());
	}
	
	public void setWindowTitle(String newtitle) {
		String wintitle = win.getTitle();
		if (!wintitle.equals(newtitle))
			win.setTitle(newtitle);
	}
	
	public void setAnimationState(double value, boolean stabilize) {
		DocumentAnimator document = getDocument();
		if (document != null) {
//			emptyAnimationQueue();
			document.setAnimationState(value);
			if (stabilize)
				getOverview().stabilize();
			getTimeline().updateCursorPosition();
			updateWindowTitle();
			repaint();
			getOverview().repaint();
		}
	}
	
	public double getOverallAnimationState() {
		if (currentAnimation == null)
			return 0;
		return currentAnimation.document.getAnimationState();
	}
	
	/**
	 * Returns the document being animated, or the last document that has been animated.
	 * @return
	 */
	public DocumentAnimator getDocument() {
		if (currentAnimation == null)
			return null;
		return currentAnimation.document;
	}
	
	public DiffOverview getOverview() {
		return overview;
	}
	
	public CompoundDiffTimeline getTimeline() {
		return timeline;
	}
	
	public void componentResized(ComponentEvent e) {
		// Using this to initialize graphics, as componentShown is never invoked
		if (settings.defaultGraphics == null) {
			Graphics2D g = (Graphics2D)getGraphics();
			settings.defaultGraphics = g;
		}
		
		// Update document widths
		settings.lineWidth = getWidth();
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		DocumentFactory.documentsWidthChanged();
		setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
		
		// Refit
		overview.moveThumbManually(0);
		overview.refitToFiles();
		timeline.setMaxDocumentHeight(overview.getMaxDocumentHeight());
		overview.repaint();
	}
	
	public float getPageX() {
		DocumentAnimator document = getDocument();
		if (document == null)
			return 0;
		return (getWidth() - scale * settings.lineWidth)/2;
	}

	/**
	 * The usual painting code.
	 */
	public void paint(Graphics g) {
		
// ExecTime.enable(true);
ExecTime.start("detail paint");
		
		DocumentAnimator document = getDocument();
		if (document == null || settings.nominalTextLayout == null)
			return;
		
		Graphics2D g2 = (Graphics2D)g;
					
		// Paint the background
		int docwidth = (int)(scale*settings.lineWidth);
		final int margin = 10; // extra margin added to the page
		int pagex0 = (int)getPageX();
		int pagex1 = pagex0 + docwidth;
		int pagey0 = (int)(Math.max(translatey, 0));
		float minpageheight = getHeight();
		float pageheight = Math.max(minpageheight, document.getCurrentHeight());
		int pagey1 = (int)(Math.min(translatey + pageheight * scale, getHeight()));
		pagey1 = Math.min(pagey1, getHeight());
		g2.setColor(BG_COLOR);
		g2.fillRect(0, 0, getWidth(), getHeight());
		g2.setColor(DOC_COLOR);
		g2.fillRect(pagex0 - margin, pagey0 - margin, pagex1 - pagex0 + margin * 2, pagey1 - pagey0 + margin * 2);
		
		// Clip
		Shape oldclip = g2.getClip();
		g2.clipRect(pagex0, 0, docwidth, getHeight());

		// Scroll & scale
		AffineTransform at0 = g2.getTransform();
		g2.translate(getPageX(), translatey);
		g2.scale(scale, scale);
				
		// Paint the text zones
		document.paint(g2, PaintMode.DETAIL, 3/scale, 3/scale);
		
		g2.setTransform(at0);
		g2.setClip(oldclip);
		
		// DEBUG
		if (WORD_DISPLACEMENT_DEBUG)
			settings.speedNormalizer.paintVisibleWordMotion(g2);
		
ExecTime.end("detail paint");
ExecTime.end("  highlight paint");
ExecTime.end("  text paint");
ExecTime.end("     drawString");

	}

	
	static public boolean deleteDirectory(File path) {
	    if( path.exists() ) {
	      File[] files = path.listFiles();
	      for(int i=0; i<files.length; i++) {
	         if(files[i].isDirectory()) {
	           deleteDirectory(files[i]);
	         }
	         else {
	           files[i].delete();
	         }
	      }
	    }
	    return( path.delete() );
	  }

	
	/**
	 * Puts this component in a window and shows it.
	 */
	public static void main(String[] args) {
		// create document window
		JFrame win = new JFrame();
		
		// Clean up the wikipedia download folder
		File download = new File(JRevisionSelector.DOWNLOAD_DIR);
		deleteDirectory(download);
		
		final DiffAnimator detail = new DiffAnimator(win);
		JPanel border1 = new JPanel();
		border1.setBackground(Color.white);
		border1.setBorder(BorderFactory.createEmptyBorder(6, 6, 6, 6));
		border1.setLayout(new BorderLayout());
		border1.add(detail, BorderLayout.CENTER);
		win.getContentPane().setLayout(new BorderLayout());
		win.getContentPane().add(border1, BorderLayout.CENTER);
		
		final DiffOverview overview = new DiffOverview(detail);
		detail.setOverview(overview);
		JPanel border2 = new JPanel();
		border2.setBackground(Color.white);
		border2.setBorder(BorderFactory.createEmptyBorder(1, 1, 1, 1));
		border2.setLayout(new BorderLayout());
		border2.add(overview, BorderLayout.CENTER);
		win.getContentPane().add(border2, BorderLayout.WEST);
		
		final CompoundDiffTimeline timeline = new CompoundDiffTimeline(detail);
		detail.setTimeline(timeline);
		JPanel border3 = new JPanel();
		border3.setBackground(Color.white);
		border3.setBorder(BorderFactory.createEmptyBorder(1, 1, 1, 1));
		border3.setLayout(new BorderLayout());
		border3.add(timeline, BorderLayout.CENTER);
		win.getContentPane().add(border3, BorderLayout.NORTH);
		
		win.pack();
		win.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		// create other objects
		detail.setSpeedNormalizer(new SpeedNormalizer(overview, detail));
		
		// create document selection dialog
		JDialog rs = new JRevisionSelector(detail, "./data", defaultFile, win);
		
		// create animation settings dialog
		JDialog as = new JAnimationSettings(detail, win);
		
		
		// center windows
		Dimension s = Toolkit.getDefaultToolkit().getScreenSize();
		win.setLocation((s.width - win.getWidth() + rs.getWidth())/2, (s.height - win.getHeight())/2);
		rs.setLocation(win.getX() - rs.getWidth() - 10, win.getY());
		as.setLocation(win.getX() - as.getWidth() - 10, win.getY()+rs.getHeight() + 10);
		
		// show windows
		as.setVisible(true);
		win.setVisible(true);
		rs.setVisible(true);
		win.requestFocus();
	}
	
	public void setTranslateY(float y) {
		translatey = y;
		repaint();
	}
	
	public void setScale(float scale) {
		this.scale = scale;
		repaint();
	}
	
	public float getTranslateY() {
		return translatey;
	}
	
	public float getScale() {
		return scale;
	}

	public void componentHidden(ComponentEvent e) {}
	public void componentMoved(ComponentEvent e) {}
	public void componentShown(ComponentEvent e) {}

	float keyScrollAmount;

	@Override
	public void keyPressedOnce(KeyEvent e) {
		switch (e.getKeyCode()) {
		case KeyEvent.VK_UP:
		case KeyEvent.VK_DOWN:
			getOverview().setStabilizationEnabled(false);
			keyScrollAmount = 2; // in pixels
			break;
		case KeyEvent.VK_HOME:
			getOverview().moveThumbToTop();
			break;
		case KeyEvent.VK_END:
			getOverview().moveThumbToBottom();
			break;
		}
	}
	
	@Override
	public void keyPressed(KeyEvent e) {
		switch (e.getKeyCode()) {
		case KeyEvent.VK_LEFT:
			getTimeline().advance(-1);
			break;
		case KeyEvent.VK_RIGHT:
			getTimeline().advance(1);
			break;
		case KeyEvent.VK_PAGE_UP:
			getOverview().pageTranslateThumb(-1);
			keyScrollAmount = keyScrollAmount * 1.05f + 0.25f;
			break;
		case KeyEvent.VK_PAGE_DOWN:
			getOverview().pageTranslateThumb(1);
			keyScrollAmount = keyScrollAmount * 1.05f + 0.25f;
			break;
		}
	}

	@Override
	public void keyReleased(KeyEvent e) {
		switch (e.getKeyCode()) {
		case KeyEvent.VK_UP:
		case KeyEvent.VK_DOWN:
			getOverview().setStabilizationEnabled(true);
			break;
		}
	}

	@Override
	public void keyRepeated(KeyEvent e) {
		switch (e.getKeyCode()) {
		case KeyEvent.VK_UP:
			getOverview().translateThumb(-keyScrollAmount);
			keyScrollAmount = keyScrollAmount * 1.05f + 0.25f;
			if (keyScrollAmount > 50)
				keyScrollAmount = 50;
			break;
		case KeyEvent.VK_DOWN:
			getOverview().translateThumb(keyScrollAmount);
			keyScrollAmount = keyScrollAmount * 1.05f + 0.25f;
			if (keyScrollAmount > 50)
				keyScrollAmount = 50;
			break;
		}
	}
	
	@Override
	public void keyTyped(KeyEvent e) {
		// TODO Auto-generated method stub
		
	}
}
