/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.diffanimation;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.Constructor;
import java.util.ArrayList;

import fr.aviz.animation.AnimatedFloat;
import fr.aviz.diff.Diff;
import fr.aviz.diff.DiffChecker;
import fr.aviz.diff.DiffUtils;
import fr.aviz.diff.Diff.Operation;
import fr.aviz.diff.methods.PonsDiff;
import fr.aviz.diffanimation.TextBlock;
import fr.aviz.diffanimation.TextBlock.Type;
import fr.aviz.utils.ExecTime;
/**
 * This class builds a two-versionned document 
 * A two-versionned document is made of a list of text blocks among
 * STABLE / INSERT / DELETE / MOVE 
 * 
 * Note that the moves are not well computed in the Diff class !!!
 * 
 * @author chevalie
 *
 */




public class Document {
	
	private String filename1, filename2;
	private String textv1;
	private String textv2;
	private Operation[] diffs;
	protected Diff diff;
	protected ArrayList<TextParagraph> paragraphs;
	public static char newline = 13;   // CR as new line
	
	/* Animation */
	private AnimatedFloat animation_coeff1 = new AnimatedFloat(0); // the one used for rendering
	private AnimatedFloat animation_coeff2 = new AnimatedFloat(0); // the one used for computing stuff
	public static enum Coeff {DEFAULT, ALT};
	protected Coeff current_animation_coeff;
	
	/* Animation settings */
	public Settings settings;

	public enum Version {
		V1,
		V2,
		V1_V2,
	}

	public Document(String filename1, String filename2, Settings settings) {
		this.filename1 = filename1;
		this.filename2 = filename2;
		this.settings = settings;

		// Compute the diffs
		textv1 = fileReader(filename1);
		textv2 = fileReader(filename2);
				
//ExecTime.enable(true);
ExecTime.start("diff");
		diff = DiffUtils.createDiff(settings.diffMethod, textv1, textv2);
ExecTime.end("diff");
ExecTime.enable(false);

		// Check if diff is correct
		if (!DiffChecker.check(textv1, textv2, diff.getDiffs())) { 
			System.err.println(diff.getClass().getSimpleName() + " is incorrect on " + this.toString());
			if (!this.filename1.contains("differrors"))
				DiffChecker.logDiffError(diff.getClass().getSimpleName() + " on " + this.toString(), textv1, textv2, diff.getDiffs());
		}
		
		//
		paragraphs = new ArrayList<TextParagraph>();
		computeParagraphs();
		
		// -- Move blocks --
		//splitMovedBlocks();
		//computeIndexesAfterMove();
	}
	
	public static String fileReader(String filename) {
		String text="";
		
		try{
			InputStream ips=new FileInputStream(filename); 
			InputStreamReader ipsr=new InputStreamReader(ips);
			BufferedReader br=new BufferedReader(ipsr);
			String line;
			while ((line=br.readLine())!=null){
				text+= line + Document.newline;
			}
			br.close(); 
		}		
		catch (Exception e){
			System.out.println(e.toString());
			return null;
		}
		
		return text;
	}
	
	public boolean equals(Document doc2) {
		return equals(doc2.filename1, doc2.filename2);
	}
	
	public boolean equals(String filename1, String filename2) {
		return (this.filename1.equals(filename1) && this.filename2.equals(filename2));
	}
	
	public ArrayList<TextParagraph> getParagraphs() {
		return paragraphs;
	}
	
	private void computeParagraphs() {
		paragraphs.clear();
		computeBlocks();
		computeLength();
		computeBlockPositions();
		
	}
	
	
	private void computeBlockPositions() {
		for (TextParagraph tp : paragraphs) {
			int pos_v1 = 0;
			int pos_v2 = 0;
			for (TextBlock tb : tp.getBlocks()) {
				tb.setIndexParagraph(Version.V1, pos_v1);
				tb.setIndexParagraph(Version.V2, pos_v2);
				pos_v1 += tb.length(Version.V1);
				pos_v2 += tb.length(Version.V2);
			}
		}
	}
	
	
	

	public void computeBlocks() {
		diffs = diff.getDiffs();
		
		int currentIdx = 0; // first character index
		int currentIdxInParagraph = 0;
		ArrayList <TextBlock> blocks = new ArrayList <TextBlock>();
		blocks.clear();
				
		TextParagraph paragraph = new TextParagraph(currentIdx);
		
		// 
		for (int i = 0 ; i < diffs.length; i++) {
			Operation op = diffs[i];	
			if (op.type == Diff.Type.MOVE) {
				System.err.println("ComputeBlocks: move operation not supported, skipped");
				continue;
			}
			
			else {
				// Stable text before the modified text
				if (op.from != 0) {
					String stableTextBefore = textv1.substring(currentIdx, op.from < currentIdx ? currentIdx : op.from);

					while (!stableTextBefore.isEmpty()) {
						int idxNewLine = stableTextBefore.indexOf(newline);
						if (idxNewLine != -1) {
							// Insert the last block of the new paragraph and create a new one
							String text = stableTextBefore.substring(0, idxNewLine);  // do not include '\n'
							//System.out.println("Flushing S-text: $$-" + text + "-$$");
							paragraph.addTextBlock(new TextBlock(currentIdx, text.length(), text, Type.STABLE, true));
							currentIdx += text.length()+1; // skip the '\n'
							paragraphs.add(paragraph);
							currentIdxInParagraph = 0;
							paragraph = new TextParagraph(currentIdx);
							stableTextBefore= stableTextBefore.substring(idxNewLine+1);
						}
						else {
							paragraph.addTextBlock(new TextBlock(currentIdx, stableTextBefore.length(), stableTextBefore, Type.STABLE, false));
							//System.out.println("End S-text:  $$-" + stableTextBefore + "-$$");
							currentIdx += stableTextBefore.length();
							currentIdxInParagraph += stableTextBefore.length();
							stableTextBefore = "";
						}
					}
				}
				
				
				// Modified part: case DELETE
				if (op.type == Diff.Type.DELETE && currentIdx + op.length <= textv1.length()) {
					String deletedText = textv1.substring(currentIdx, currentIdx + op.length);
					//System.err.println("D-TO-FLUSH: " + deletedText);
					
					while (!deletedText.isEmpty()) {
						//System.out.println("Deleted text: " + deletedText);
						int idxNewLine = deletedText.indexOf(newline);
						if (idxNewLine != -1) {
							// Insert the last block of the new paragraph and create a new one
							String text = deletedText.substring(0, idxNewLine);
							paragraph.addTextBlock(new TextBlock(currentIdx, text.length(), text, Type.DELETE, true));
							currentIdx += text.length()+1;
							paragraphs.add(paragraph);
							currentIdxInParagraph = 0;
							paragraph = new TextParagraph(currentIdx);
							deletedText = deletedText.substring(idxNewLine+1);
						}
						else {
							paragraph.addTextBlock(new TextBlock(currentIdx, deletedText.length(), deletedText, Type.DELETE, false));
							currentIdx += deletedText.length();
							currentIdxInParagraph += deletedText.length();
							deletedText = "";
						}
					}	
				}
				
				// Modified part: case INSERT
				//else 
				if (op.type == Diff.Type.INSERT) {
					String insertedText = op.contents;
					//System.out.println("I-Text: " + insertedText);
					
					int currentIdxInsertion = currentIdx;
					//paragraphBeforeInsertion = paragraph;
					while (!insertedText.isEmpty()) {
						int idxNewLine = insertedText.indexOf(newline);
						if (idxNewLine != -1) {
							// Insert the last block of the new paragraph and create a new one
							String text = new String(insertedText.substring(0, idxNewLine));
							//System.out.println("Flush I-text: $$-" + text + "-$$");
							paragraph.addTextBlock(new TextBlock(currentIdxInsertion, text.length(), text, Type.INSERT, true));
							//currentIdxInsertion += text.length()+1;
							paragraphs.add(paragraph);
							paragraph = new TextParagraph(currentIdxInsertion);
							insertedText = insertedText.substring(idxNewLine+1);
							//System.out.println("New line inserted = true");
						}
						else {
							paragraph.addTextBlock(new TextBlock(currentIdxInsertion, insertedText.length(), insertedText, Type.INSERT, false));
							//System.out.println("End I-text: $$-" + insertedText + "-$$");
							//currentIdxInsertion += insertedText.length();
							//currentIdxInParagraphInsertion += insertedText.length();
							insertedText= "";
						}
					}	
				}
			}
		}
	
		
	
		// End of the text
		String remainingText = textv1.substring(currentIdx);
		while (remainingText.length() != 0) {
			int idxNewLine = remainingText.indexOf(newline);
			if (idxNewLine != -1) {
				// Insert the last block of the new paragraph and create a new one
				String text = remainingText.substring(0, idxNewLine);
				//System.out.println("Flush S-text: " + 0 + " - " + idxNewLine + " $$-" + text + "-$$");
				
				paragraph.addTextBlock(new TextBlock(currentIdx, text.length(), text, Type.STABLE, true));
				currentIdx += text.length()+1;
				paragraphs.add(paragraph);
				currentIdxInParagraph = 0;
				paragraph = new TextParagraph(currentIdx);
				remainingText = remainingText.substring(idxNewLine+1);
			}
			else {
				paragraph.addTextBlock(new TextBlock(currentIdx, remainingText.length(), remainingText, Type.STABLE, false));
				remainingText = "";
			}
		}
	}
	
	
	/**
	 * Compute the positions of each text paragraph at version 2
	 */
	private void computeLength() {
		for (TextParagraph tp : paragraphs) {
			tp.computeLength_v2();
		}
	}
	
/*	public void setVersion(int version) {
		animation_coeff.directSet(version == V2 ? 1 : 0);	
	*/
	protected void maybeUpdateGeometry() {}

	protected AnimatedFloat animation_coeff() {
		return current_animation_coeff == Coeff.DEFAULT ? animation_coeff1 : animation_coeff2;
	}
	
	public void setAnimationCoeff(Coeff coeff) {
		current_animation_coeff = coeff;
	}
	
	public void startAnimation(float t0, float t1) {
		if (settings.animationEnabled && !filename1.equals(filename2)) {
			animation_coeff().directSet(t0);
			animation_coeff().set(t1);
		} else {
//			animation_coeff.directSet(t0);
			maybeUpdateGeometry();
			animation_coeff().directSet(t1);
		}
		maybeUpdateGeometry();
	}
	
	public void changeAnimation(float t1) {
		animation_coeff().set(t1);
	}
	
	public void endAnimation() {
		animation_coeff().directSet(animation_coeff().directGet());
		AnimatedFloat.animateAll();
	}
	
	public String toString() {
		return filename1 + " <-> " + filename2;
	}

	public String getFilename1() {
		return filename1;
	}
	
	public String getFilename2() {
		return filename2;
	}
}
