/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.diffanimation;

import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.font.TextLayout;
import java.util.ArrayList;

import fr.aviz.diff.Diff;
import fr.aviz.diff.DiffUtils;
import fr.aviz.diff.Diff.Operation;
import fr.aviz.diffanimation.TextBlock.TextWord;
import fr.aviz.diffanimation.TextBlock.Type;
import fr.aviz.diffanimation.render.TextRenderer;
import fr.aviz.utils.WordAdvanceCache;

/**
 * 
 * This class does the animation of a document.
 * 
 * The idea is to progressively show/hide the text that is inserted/deleted 
 * by letting/shrinking room for the modified text.
 * 
 * @author chevalie
 *
 */

public class DocumentAnimator extends Document {
	
	/**
	 * The extra space added to the bottom of the document, expressed in pixels
	 */
	public static float BOTTOM_DOCUMENT_MARGIN = 8;
	public static boolean STDOUT_DEBUG = false;
	
	public enum PaintMode {DETAIL, OVERVIEW, HISTORYFLOW};

	private TextRenderer textRenderer;
	
	float motionPause;

	boolean turbo;
	
	private Version v_start;
	private Version v_end;
	
	private boolean startEndGeometryDirty = true;
	private boolean wordSizesDirty = true;

	private float currentMotionAnimationState = -1f;
	private float currentDocumentHeight = 0;
	
	private float changeMagnitude;
	
	// Memory optimization
	float []origin = new float[]{0,0};
	float []pupleft_vstart = new float[]{0, 0};
	float []pdownright_vstart = new float[]{0, 0};
	float []pupleft_vend = new float[]{0, 0};
	float []pdownright_vend = new float[]{0, 0};
	float []pupleft_motion = new float[]{0,0};
	float []pdownright_motion = new float[]{0,0};
	
	float []bupleft_vstart = new float[]{0, 0};
	float []bdownright_vstart = new float[]{0, 0};
	float []bupleft_vend = new float[]{0, 0};
	float []bdownright_vend = new float[]{0, 0};
	float []bupleft_motion = new float[]{0,0};
	float []bdownright_motion = new float[]{0,0};
	float []bupleft_directmotion = new float[]{0,0};
	float []bdownright_directmotion = new float[]{0,0};
	float []bupleft_modulomotion = new float[]{0,0};
	float []bdownright_modulomotion = new float[]{0,0};
	
	float []wupleft_vstart = new float[]{0, 0};
	float []wdownright_vstart = new float[]{0, 0};
	float []wupleft_vend = new float[]{0, 0};
	float []wdownright_vend = new float[]{0, 0};
	float []wupleft_motion = new float[]{0,0};
	float []wdownright_motion = new float[]{0,0};
	
	
	float []loc_v1 = new float[]{0,0};
	float []loc_v2 = new float[]{0,0};
	
	
	
	public DocumentAnimator(String v1, String v2, Settings settings) {
		super(v1, v2, settings);
		computeChangeMagnitude();
		//setStartEndVersion(DiffAnimator.DEFAULT_START, DiffAnimator.DEFAULT_END);
		setStartEndVersion(settings.vstart, settings.vend);
		textRenderer = new TextRenderer(this);
		diff = null; // release memory
	}
	
	/**
	 * This boosts the animation in case it's in the middle of a sequence.
	 * @param turbo
	 */
	public void setTurbo(boolean turbo) {
		this.turbo = turbo;
	}
	
	public boolean getTurbo() {
		return turbo;
	}
	
	public void setAnimationState(double value) {
		animation_coeff().directSet(value);
	}
	
	public void markWordSizesDirty() {
		wordSizesDirty = true;
	}
	
	public void markStartEndGeometryDirty() {
		startEndGeometryDirty = true;
	}
	
	public void setStartEndVersion(Version v_start, Version v_end) {
		if (v_start == v_end || v_start == Version.V2 || v_end == Version.V1) {
			System.err.println("DocumentAnimator:setStartEndVersion error, invalid versions");
			return;
		}
		this.v_start = v_start;
		this.v_end = v_end;
		
		// Update all references
		for (TextParagraph tp : paragraphs) {
			tp.setStartGeometry(v_start);
			tp.setEndGeometry(v_end);
			tp.setMotion(v_start, v_end);
			
			for (TextBlock tb : tp.getBlocks()) {
				tb.setStartGeometry(v_start);
				tb.setEndGeometry(v_end);
				tb.setMotion(v_start, v_end);
				tb.setWrap(v_start, v_end);
				
				for (TextWord tw : tb.getWords()) {
					tw.setStartGeometry(v_start);
					tw.setEndGeometry(v_end);
					tw.setMotion(v_start, v_end);
				}
			}
		}
	}
	
	public Version getStartVersion() {
		return v_start;
	}
	
	public Version getEndVersion() {
		return v_end;
	}
	
	/**
	 * Returns a float between 0 and 1 that specifies the advancement of the text
	 * motion animation. 0 is the original document, 1 is the final document.
	 * Text motion animation starts slightly later than the overall animation and
	 * ends earlier.
	 */
	public float getMotionAnimationState() {
		float t = (float)animation_coeff().get();
		if (t <= motionPause / 2)
			return 0;
		if (t >= 1 - motionPause / 2)
			return 1;
		return (t - motionPause / 2) / (1 - motionPause);
	}
	
	/**
	 * Returns a float between 0 and 1 that specifies the advancement of the text
	 * color / highlight animation (same as animation_coeff). 
	 */
	public float getColorAnimationState() {
		return getAnimationState();
	}
	
	/**
	 * Returns the overall animation state. Don't use this to animate colors or motion!!!!
	 * Use the two above methods instead.
	 */
	public float getAnimationState() {
		float t = (float)animation_coeff().get();
		if (t < 0) t = 0;
		if (t > 1) t = 1;
		return t;
	}
	
	public boolean animating() {
		return animation_coeff().get() != animation_coeff().directGet();
	}
	
	public boolean changed() {
		return animation_coeff().changed();
	}

	public void setFontContext(Graphics2D g) {
		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);	
		g.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
		//g.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);
		g.setFont(Settings.TEXT_FONTS[settings.textFontIndex]);
	}
	
	public void paint(Graphics2D g, PaintMode paintMode, float minHighlightWidth, float minHighlightHeight) {
		maybeUpdateGeometry();
		setFontContext(g);
		textRenderer.paint(g, getMotionAnimationState(), (getColorAnimationState()), paintMode, minHighlightWidth, minHighlightHeight);
	}
	
	boolean painting = false;
		
	protected void maybeUpdateGeometry() {
		// FIXME: to remove
		v_start = settings.vstart;
		v_end = settings.vend;
		
		
		boolean startEndGeometryChanged = maybeUpdateStartEndGeometry();
		float t = getMotionAnimationState();
		if (startEndGeometryChanged || currentMotionAnimationState != t) {
			updateGeometry(t);
			currentMotionAnimationState = t;
		}
	}
	
	
	public void updateGeometry(float motion_t) {
		currentDocumentHeight = textRenderer.updateGeometry(motion_t);
	}
	
	private boolean maybeUpdateStartEndGeometry() {
		if (settings.defaultGraphics == null)
			return false;
		if (startEndGeometryDirty) {
			updateStartEndGeometry(settings.defaultGraphics);
			startEndGeometryDirty = false;
			return true;
		}
		return false;
	}
	
	private void updateStartEndGeometry(Graphics2D g) {
		maybeUpdateWordSizes(g);
		precomputeGeometry(Version.V1);
		precomputeGeometry(Version.V2);
		precomputeGeometry(Version.V1_V2);
		precomputeMotions(Version.V1, Version.V2);
		precomputeMotions(Version.V1, Version.V1_V2);
		precomputeMotions(Version.V1_V2, Version.V2);
	}
	
	private void maybeUpdateWordSizes(Graphics2D g) {
		if (wordSizesDirty) {
			computeWordSizes(g);
			wordSizesDirty = false;
		}
	}
	
	/**
	 * Returns the text block at the given y-coordinate, or the block just above if this
	 * is an empty area.
	 * 
	 * @param y
	 * @return
	 */
	public TextBlock getTextBlockAt(float y) {
		
		maybeUpdateGeometry();

		for (TextParagraph tp : getParagraphs()) {
			for (TextBlock tb : tp.getBlocks()) {
				if (tb.getGeometry()[1] >= y - 2 && tb.getText().trim().length() > 0 && tb.getScalex() > TextBlock.MIN_SCALE)
					return tb;
			}
		}
		
		// return the last block.
		TextParagraph lastpar = getParagraphs().get(getParagraphs().size()-1);
		return lastpar.getBlocks().get(lastpar.getBlocks().size()-1);
	}
	
	/**
	 * Returns the top y-coordinate of the text block
	 * 
	 * @param tb
	 * @return
	 */
	public float getTextBlockTop(TextBlock tb) {
		maybeUpdateGeometry();
//		if (!containsTextBlock(tb))
//			return -1;
		return tb.getGeometry()[1];
	}
	
	/**
	 * Returns whether the document contains the TextBlock object.
	 * 
	 * @param tb_
	 * @return
	 */
	public boolean containsTextBlock(TextBlock tb_) {
		for (TextParagraph tp : getParagraphs()) {
			for (TextBlock tb : tp.getBlocks()) {
				if (tb == tb_)
					return true;
			}
		}	
		return false;
	}
	
	/**
	 * Returns the bottom y-coordinate of the text block
	 * 
	 * @param tb
	 * @return
	 */
	public float getTextBlockBottom(TextBlock tb) {
		maybeUpdateGeometry();
		return tb.getGeometry()[3];
	}
	
	/**
	 * Returns all words between y0 and y1. If skipInvisible = true, will skip
	 * all zero-length words and words for which scalex = 0.
	 */
	public ArrayList<TextWord> getAllWords(float y0, float y1, boolean skipInvisible) {
		ArrayList<TextWord> words = new ArrayList<TextWord>();
		float [] geom;
		float ymid;
		for (TextParagraph tp : getParagraphs()) {
			geom = tp.getGeometry();
			if (geom[3] >= y0 && geom[1] <= y1) {
				for (TextBlock tb : tp.getBlocks()) {
					geom = tb.getGeometry();
					if (geom[3] >= y0 && geom[1] <= y1) {
						for (TextWord tw : tb.getWords()) {
							geom = tw.getGeometry();
							ymid = (geom[1] + geom[3])/2;
							if (ymid > y0 && ymid < y1) {
								if (!skipInvisible || (tw.getText().length() > 0 && geom[2] > geom[0]))
									words.add(tw);
							}
						}
					}
				}
			}
		}
		return words;
	}
	
	/**
	 * Returns the bottom y-coordinate of the document.
	 * @return
	 */
	public float getCurrentHeight() {
		maybeUpdateGeometry();
		return currentDocumentHeight + BOTTOM_DOCUMENT_MARGIN;
	}
	
	public float getHeight(float t) {
		
		if (t == getAnimationState()) {
			maybeUpdateGeometry();
			return getCurrentHeight();
		} else {
			setAnimationCoeff(DocumentAnimator.Coeff.ALT);
			setAnimationState(t);
			maybeUpdateGeometry();
			float height = getCurrentHeight();
			setAnimationCoeff(DocumentAnimator.Coeff.DEFAULT);
			return height;
		}
	}
	
	/**
	 * This is used by the timeline.
	 * For now, this function just returns the natural logarithm of the total
	 * number of characters inserted and deleted.
	 */
	private void computeChangeMagnitude() {
		changeMagnitude = DiffUtils.getEditDistance(diff.getDiffs());
		changeMagnitude = (float)Math.log(1 + changeMagnitude);
	}
	
	public float getChangeMagnitude() {
		return changeMagnitude;
	}
	
	private void maybeUpdateFontData(Graphics2D g) {
		if (settings.nominalTextLayout == null) {
			setFontContext(g);
			settings.fontRenderContext = g.getFontRenderContext();
			settings.nominalTextLayout = new TextLayout("Âj", Settings.TEXT_FONTS[settings.textFontIndex], settings.fontRenderContext);
			settings.lineHeight = (int)Math.floor((settings.nominalTextLayout.getAscent() + settings.nominalTextLayout.getDescent() + settings.nominalTextLayout.getLeading()) * settings.LINE_SPACING);
			settings.leading = settings.lineHeight - (settings.nominalTextLayout.getAscent() + settings.nominalTextLayout.getDescent());
			settings.baseline = settings.nominalTextLayout.getAscent() + settings.leading / 2.f;
			settings.ascent = settings.nominalTextLayout.getAscent();
		}
	}

	private void computeWordSizes(Graphics2D g) {
		maybeUpdateFontData(g);
//WordAdvanceCache.enable(false);
//ExecTime.start("Word lengths");
		computePartialWordLengths();
		computeFullWordLengths(Version.V1);
		computeFullWordLengths(Version.V2);
		computeFullWordLengths(Version.V1_V2);
//ExecTime.pause("Word lengths");
	}
	
	private void computePartialWordLengths() {
		String word;
		for (TextParagraph tp : paragraphs) {
			for (TextBlock tb : tp.getBlocks()) {
				for (TextWord tw : tb.getWords()) {
					word = tw.getText();
					if (word.length() == 0) {
						// set the length of the word
						tw.setAdvance(0);
					} else { 
						// set the length of the word
						tw.setAdvance(WordAdvanceCache.getAdvance(settings, word));
						tw.setVisibleAdvance(WordAdvanceCache.getVisibleAdvance(settings, word));
					}
				}
			}
		}
	}

	/**
	 * Used for line breaks.
	 * 
	 * Actual words are sometimes split out across several blocks. This retrieves the whole
	 * words for the given version and computes their extent (trailing spaces not included).
	 * Whole word extents are stored in the first TextWord.
	 */
	private void computeFullWordLengths(Version version) {
		TextBlock.Type ignored = null;
		
		// Depending on the animation, the ignored version differs
		if (version == Version.V1)
			ignored = TextBlock.Type.INSERT;
		else if (version == Version.V2)
			ignored = TextBlock.Type.DELETE;
		
		String word, fullword = null;
		TextBlock tb;
		TextWord tw = null, fullwordstart = null;
		
		for (TextParagraph tp : paragraphs) {
			
			int tb_count = tp.getBlocks().size();
			for (int tbi=0; tbi<tb_count; tbi++) {
				
				tb = tp.getBlocks().get(tbi);
				boolean ignore = tb.getType() == ignored;
				int tw_count = tb.getWords().size();
				for (int twi=0; twi<tw_count; twi++) {
					
					if (!ignore) {
						tw = tb.getWords().get(twi);
						word = tw.getText();
						if (fullword == null) {
							fullword = word;
							fullwordstart = tw;
						} else
							fullword += word;
						
						tw.setFullWordAdvance(0, version);
					}
					
					boolean endOfWord = !ignore && tw.getTrailingSpaces() > 0;
					boolean lastWordInParagraph = (tbi == tb_count-1 && twi == tw_count-1);
					if (fullword != null && (endOfWord || lastWordInParagraph)) {
						// now we've reached the end of the whole word, set its length on the first 
						// partial word composing it.
						if (fullword.length() == 0)
							fullwordstart.setFullWordAdvance(0, version);
						else {
							fullwordstart.setFullWordAdvance(WordAdvanceCache.getVisibleAdvance(settings, fullword), version);
						}
						fullword = null;
					}
				}
			}
		}
	}
	



	private float vectorLength(float []v) {
		return (float) Math.sqrt(v[0]*v[0] + v[1]*v[1]);
	}
	
	/**
	 * Returns the local position of a vector
	 * @param v a coordinate vector {x1,y1,x2,y2,...,xn,yn}
	 * @param vref the coordinates of the reference point {x_ref, y_ref}
	 * @return the local coordinates
	 */
	private void computeLocalGeometry(float []v, float[]vref, float []vloc) {
		vloc[0] = v[0] - vref[0];
		vloc[1] = v[1] - vref[1];
		
		/*
		if (v.length % 2 != 0 || vref.length < 2) {
			System.err.println("DocumentAnimator:computeLocalGeometry error coordinates");
			return null;
		}
		
		float []vloc = new float[v.length];
		for (int i = 0; i < v.length; i+=2) {
			vloc[i] = v[i] - vref[0];
			vloc[i+1] = v[i+1] - vref[1];
		}
		return vloc;*/
	}
	
	
	/**
	 * Returns the direct interpolation of two coordinates vectors,
	 * relatively to a reference point
	 * @param v1
	 * @param v2
	 * @param motion_ref  the motion vector of the reference point
	 * @return direct motion of a point
	 */
	private void directInterpolation(float []v1, float []v2, float []v_ref_v1, float []v_ref_v2, float []motion_ref, float []motion) {
		/*if (v1.length != v2.length || v1.length % 2 != 0 || motion_ref.length % 2 != 0) {
			System.err.println("DocumentAnimator:directInterpolation error coordinates");
			return null;
		}*/
		
		computeLocalGeometry(v1, v_ref_v1, loc_v1);
		computeLocalGeometry(v2, v_ref_v2, loc_v2);
			
		motion[0] = motion_ref[0] + (loc_v2[0]-loc_v1[0]);
		motion[1] = motion_ref[1] + (loc_v2[1]-loc_v1[1]);
		
		/*float v[] = new float[v1.length];
		for (int i = 0 ; i < v1.length ; i+=2) {
			v[i] = motion_ref[0] + (loc_v2[i]-loc_v1[i]);
			v[i+1] = motion_ref[1] + (loc_v2[i+1]-loc_v1[i+1]);
		}
		return v;*/
	}
	
	/**
	 * Returns the direct interpolation of two coordinates vectors,
	 * relatively to a reference point
	 * The modulo is always lineWidth, because we never consider modulo
	 * positions when a block expands / shrinks
	 * @param v1
	 * @param v2
	 * @param motion_ref  the motion vector of the reference point
	 * @param yOffset  used to consider the same line basis
	 * @return direct motion of a point
	 */
	private void moduloInterpolation(float []v1, float []v2, float []v_ref_v1, float [] v_ref_v2, float []motion_ref, float yOffset_v1, float yOffset_v2, float []motion) {
		/*if (v1.length != v2.length || v1.length / 2 != 1 || motion_ref.length % 2 != 0) 
			return null;*/
		
		float w = settings.lineWidth;
		float h = settings.lineHeight;
		computeLocalGeometry(v1, v_ref_v1, loc_v1);
		computeLocalGeometry(v2, v_ref_v2, loc_v2);
		
		// Put everything on the first line
		while (loc_v1[1]-yOffset_v1 > 0) { 
			loc_v1[0] += w;
			loc_v1[1] -= h;
		}
		while (loc_v2[1]-yOffset_v2 > 0) {
			loc_v2[0] += w;
			loc_v2[1] -= h;
		}
		
		motion[0] = motion_ref[0] + (loc_v2[0]-loc_v1[0]);
		motion[1] = motion_ref[1] + (loc_v2[1]-loc_v1[1]);
		
		/*
		float v[] = new float[v1.length];
		for (int i = 0 ; i < v1.length ; i+=2) {
			v[i] = motion_ref[0] + (loc_v2[i]-loc_v1[i]);
			v[i+1] = motion_ref[1] + (loc_v2[i+1]-loc_v1[i+1]);
		}
		return v;*/
	}
	
	
	private void precomputeMotions(Version vstart, Version vend) {
		for (TextParagraph tp : paragraphs) {
			// -- 1 -- Compute paragraph motion
			// A paragraph has always a direct motion (up left) and a modulo motion (down right)
			float pgeom_vstart[] = tp.getPrecomputedGeometry(vstart);
			float pgeom_vend[] = tp.getPrecomputedGeometry(vend);
			
			pupleft_vstart[0] = pgeom_vstart[0];
			pupleft_vstart[1] = pgeom_vstart[1];
			pdownright_vstart[0] = pgeom_vstart[2];
			pdownright_vstart[1] = pgeom_vstart[3];
			pupleft_vend[0] = pgeom_vend[0];
			pupleft_vend[1] = pgeom_vend[1];
			pdownright_vend[0] = pgeom_vend[2];
			pdownright_vend[1] = pgeom_vend[3];
			
			// here test paragraph height if it equals zero: no offset to compute wrapped position
			float yOffset_vstart = 0;
			float yOffset_vend = 0;
			float pheight_vstart = pdownright_vstart[1] - pupleft_vstart[1];
			float pheight_vend = pdownright_vend[1] - pupleft_vend[1];
			
			if (pheight_vstart == 0 && pheight_vend != 0)
				yOffset_vend = settings.lineHeight;
			if (pheight_vstart != 0 && pheight_vend == 0)
				yOffset_vstart = settings.lineHeight;
				
			directInterpolation(pupleft_vstart, pupleft_vend, origin, origin, origin, pupleft_motion);
			moduloInterpolation(pdownright_vstart, pdownright_vend, pupleft_vstart, pupleft_vend, pupleft_motion, yOffset_vstart, yOffset_vend, pdownright_motion);			
			
			tp.setPrecomputedMotion(vstart, vend, pupleft_motion[0], pupleft_motion[1], pdownright_motion[0], pdownright_motion[1]);
			
			
			for (TextBlock tb : tp.getBlocks()) {
				// -- 2 -- Compute block motion
  				float bgeom_vstart[] = tb.getPrecomputedGeometry(vstart);
  				float bgeom_vend[] = tb.getPrecomputedGeometry(vend);
  				
  				
  				bupleft_vstart[0] = bgeom_vstart[0];
  				bupleft_vstart[1] = bgeom_vstart[1];
  				bdownright_vstart[0] = bgeom_vstart[2];
  				bdownright_vstart[1] = bgeom_vstart[3];
  				bupleft_vend[0] = bgeom_vend[0];
  				bupleft_vend[1] = bgeom_vend[1];
  				bdownright_vend[0] = bgeom_vend[2];
  				bdownright_vend[1] = bgeom_vend[3];
  				
  				float bup_yOffset_vstart = 0;
  				float bup_yOffset_vend = 0;
  				float bdown_yOffset_vstart = 0;
  				float bdown_yOffset_vend = 0;
  				float bheight_vstart = bdownright_vstart[1] - bupleft_vstart[1];
  				float bheight_vend = bdownright_vend[1] - bupleft_vend[1];
  				
  				if (bheight_vstart == 0 && bheight_vend != 0) {
  					bup_yOffset_vstart = settings.lineHeight/2f;
  					bdown_yOffset_vend = settings.lineHeight;
  				}
  				if (bheight_vstart != 0 && bheight_vend == 0) {
  					bup_yOffset_vend = settings.lineHeight/2f;
  					bdown_yOffset_vstart = settings.lineHeight;
  				}
  				
  				
  				
  				// A block has 3 different kind of motion 
  				// - If upleft corner wraps, downright corner wrap  --> all wrap
  				// - If upleft corner moves directly, downright corner may move directly (expands / shrinks)
  				//                                                         wrap

  				directInterpolation(bupleft_vstart, bupleft_vend, pupleft_vstart, pupleft_vend, pupleft_motion, bupleft_directmotion);
  				moduloInterpolation(bupleft_vstart, bupleft_vend, pupleft_vstart, pupleft_vend, pupleft_motion, bup_yOffset_vstart, bup_yOffset_vend, bupleft_modulomotion);
  				
  				if (vectorLength(bupleft_directmotion) > vectorLength(bupleft_modulomotion)) {
  					tb.setPrecomputedWrapUpLeft(vstart, vend, true);
  					tb.setPrecomputedWrapDownRight(vstart, vend, true);
  					bupleft_motion = bupleft_modulomotion;
  					moduloInterpolation(bdownright_vstart, bdownright_vend, bupleft_vstart, bupleft_vend, bupleft_motion, bdown_yOffset_vstart, bdown_yOffset_vend, bdownright_motion);
  				} 
  				else {
  					tb.setPrecomputedWrapUpLeft(vstart, vend, false);
  					bupleft_motion = bupleft_directmotion;
  				
  					moduloInterpolation(bdownright_vstart, bdownright_vend, bupleft_vstart, bupleft_vend, bupleft_motion, bdown_yOffset_vstart, bdown_yOffset_vend, bdownright_modulomotion);
  					directInterpolation(bdownright_vstart, bdownright_vend, bupleft_vstart, bupleft_vend, bupleft_motion, bdownright_directmotion);
						
  					if (tb.type == Type.STABLE || //tb.type == stable ||
  						(tb.type == Type.DELETE && vend == Version.V1_V2) ||
  						(tb.type == Type.INSERT && vstart == Version.V1_V2) ||
  						vectorLength(bdownright_modulomotion) < vectorLength(bdownright_directmotion)) {
  						tb.setPrecomputedWrapDownRight(vstart, vend, true);
  						bdownright_motion = bdownright_modulomotion;
  					}
  					else {
  						tb.setPrecomputedWrapDownRight(vstart, vend, false);
  						bdownright_motion = bdownright_directmotion;	
  					}
  				}
  				
  				tb.setPrecomputedMotion(vstart, vend, bupleft_motion[0], bupleft_motion[1], bdownright_motion[0], bdownright_motion[1]);
  				
  				
  				/*
  				if (tb.getType() != Type.STABLE && vstart == Version.V1 && vend == Version.V1_V2) {
  					System.out.println("-------------------------------------------------");
  					System.out.println(tb.getText());
  					System.out.println(tb.getType());
  					System.out.println(tb.getWrapUpLeft(vstart, vend) + "  " + tb.getWrapDownRight(vstart, vend));
  				
  					System.out.println("-- Upleft");
  					System.out.println("Motion direct: " + bupleft_directmotion[0] + " " + bupleft_directmotion[1]);
  					System.out.println("Motion modulo: " + bupleft_modulomotion[0] + " " + bupleft_modulomotion[1]);
  					System.out.println("Geometry debut: " + bupleft_vstart[0] + " (" + bgeom_vstart[0] + ")" + bupleft_vstart[1] + " (" + bgeom_vstart[1] + ")");
  					System.out.println("Geometry  fin : " + bupleft_vend[0] + " (" + bgeom_vend[0] + ")" + bupleft_vend[1] + " (" + bgeom_vend[1] + ")");
  				
  					System.out.println("-- Downright");
  					System.out.println("Motion direct: " + bdownright_directmotion[0] + " " + bdownright_directmotion[1] + "   " + vectorLength(bdownright_directmotion));
  					System.out.println("Motion modulo: " + bdownright_modulomotion[0] + " " + bdownright_modulomotion[1] + "   " + vectorLength(bdownright_modulomotion));
  					System.out.println("Geometry debut: " + bdownright_vstart[0] + " (" + bgeom_vstart[2] + ")" + bdownright_vstart[1] + " (" + bgeom_vstart[3] + ")");
  					System.out.println("Geometry  fin : " + bdownright_vend[0] + " (" + bgeom_vend[2] + ")" + bdownright_vend[1] + " (" + bgeom_vend[3] + ")");
  				}
  				*/
  				
  				
  				
  				
  				// -- 3 -- Compute word motion		
  				for (TextWord tw : tb.getWords()) {
  					float wgeom_vstart[] = tw.getPrecomputedGeometry(vstart);
  	  				float wgeom_vend[] = tw.getPrecomputedGeometry(vend);
  	  	
  	  				wupleft_vstart[0] = wgeom_vstart[0];
  	  				wupleft_vstart[1] = wgeom_vstart[1];
  	  				wdownright_vstart[0] = wgeom_vstart[2];
  	  				wdownright_vstart[1] = wgeom_vstart[3];
  	  				wupleft_vend[0] = wgeom_vend[0];
  	  				wupleft_vend[1] = wgeom_vend[1];
  	  				wdownright_vend[0] = wgeom_vend[2];
  	  				wdownright_vend[1] = wgeom_vend[3];
  				
  	  				float wdown_yOffset_vstart = 0;
  	  				float wdown_yOffset_vend = 0;
  	  				float wheight_vstart = wdownright_vstart[1] - wupleft_vstart[1];
  	  				float wheight_vend = wdownright_vend[1] - wupleft_vend[1];
  				
  	  				if (wheight_vstart == 0 && wheight_vend != 0) 
  	  					wdown_yOffset_vend = settings.lineHeight;
  	  				if (wheight_vstart != 0 && wheight_vend == 0) 
  	  					wdown_yOffset_vstart = settings.lineHeight;
  	  				
  					// A word has a direct motion only if block is all direct motion (case expand / shrink) <=> !tb.wrapDownRight
  	  				
  	  				if (tb.getPrecomputedWrapDownRight(vstart, vend)) {
  	  					moduloInterpolation(wupleft_vstart, wupleft_vend, bupleft_vstart, bupleft_vend, bupleft_motion, 0, 0, wupleft_motion);
  						moduloInterpolation(wdownright_vstart, wdownright_vend, wupleft_vstart, wupleft_vend, wupleft_motion, wdown_yOffset_vstart, wdown_yOffset_vend, wdownright_motion);
  	  				}
  	  				else {
  	  					directInterpolation(wupleft_vstart, wupleft_vend, bupleft_vstart, bupleft_vend, bupleft_motion, wupleft_motion);
  	  					directInterpolation(wdownright_vstart, wdownright_vend, wupleft_vstart, wupleft_vend, wupleft_motion, wdownright_motion);
  	  				}
  	  				
  	  				tw.setPrecomputedMotion(vstart, vend, wupleft_motion[0], wupleft_motion[1], wdownright_motion[0],wdownright_motion[1]);
  				}
			}
		}
	}
	              

	
	private void precomputeGeometry(Version version) {
		// for now, only justify non-fixed fonts
		boolean justify = settings.textFontIndex != 0;		
		TextBlock.Type ignored = null;
		// Depending on the animation, the ignored version differs
		if (version == Version.V1) 
			ignored = Type.INSERT;
		else if (version == Version.V2) 
			ignored = Type.DELETE;
		
		// HERE COMPUTE geometry_V1, geometry_V2
		float currentBaseLine = 0;   // y coordinate
		float currentAdvance = 0;    // x coordinate
		
		ArrayList<TextWord> current_line = new ArrayList<TextWord>();
		int total_spaces;

		for (TextParagraph tp : paragraphs) {
			float currentParBaseLine = currentBaseLine;
			float currentParAdvance = currentAdvance;

			current_line.clear();
			total_spaces = 0;
			
			for (TextBlock tb : tp.getBlocks()) {
				float currentBlockBaseLine = currentBaseLine;
				float currentBlockAdvance = currentAdvance;
				float yOffset = 0;
				if (tb.getType() != ignored && tb.length() != 0) {
					float wrappingWidth = settings.lineWidth - currentAdvance;
					
					for (TextWord tw : tb.getWords()) {
						
						float advance = tw.getAdvance();
						float fullword_advance = tw.getFullWordAdvance(version);
						// Check line break: if this word starts a whole word (>0), look at
						// the whole word length to see if it fits.
						
						if (fullword_advance > 0) {
							// If the word does not fit...
							if (fullword_advance > wrappingWidth) {
								// 1 -- Justify all the previous words on the line
								if (justify) {
									if (current_line.size() > 1) {
										TextWord lastword = current_line.get(current_line.size() - 1);
										int inbetween_spaces = total_spaces - lastword.getTrailingSpaces();
										float extraspace = wrappingWidth / inbetween_spaces;
										int spaces_before = 0;
										for (int j=0; j<current_line.size(); j++) {
											current_line.get(j).translatex(extraspace * spaces_before, version);
											spaces_before += current_line.get(j).getTrailingSpaces();
										}
									}	
								}
								// 2 -- Start a new line
								currentBaseLine += settings.lineHeight;
								currentAdvance = 0;
								wrappingWidth = settings.lineWidth;
								current_line.clear();
								total_spaces = 0;
							}
						}
							
						// Add this word to the list of words in the line,
						// for further justification
						if (justify) {
							current_line.add(tw);
							total_spaces += tw.getTrailingSpaces();
						}
	
						//In case the word fits without the last separator, consider the visibleadvance
						if  (currentAdvance + advance > settings.lineWidth)
							advance = tw.getVisibleAdvance();
						
						wrappingWidth -= advance;
						currentAdvance += advance;
					
						tw.setPrecomputedGeometry(currentAdvance-advance, currentBaseLine, currentAdvance, currentBaseLine+settings.lineHeight, version);						
					} 
				}
				else {
					yOffset = settings.lineHeight/2.f;
					for (TextWord tw : tb.getWords()) {
						tw.setPrecomputedGeometry(currentAdvance, currentBaseLine+yOffset, currentAdvance, currentBaseLine+yOffset, version);
					}
				}
				
				tb.setPrecomputedGeometry(currentBlockAdvance, currentBlockBaseLine+yOffset, currentAdvance, currentBaseLine+settings.lineHeight-yOffset, version);
				currentBlockBaseLine = currentBaseLine;
				currentBlockAdvance = currentAdvance;
			}
			
			if (justify) {
				// normalize blocks
				for (TextBlock tb : tp.getBlocks()) {
					tb.correctPrecomputedGeometry(version);
				}
			}
			
			float p_yOffset = 0;
			if (tp.length(version) == 0)
				p_yOffset = settings.lineHeight/2.f;
			tp.setPrecomputedGeometry(currentParAdvance, currentParBaseLine+p_yOffset, currentAdvance, currentBaseLine+settings.lineHeight-p_yOffset, version);
				
			//if (DiffAnimator.ANIM_VERSION == AnimVersions.V1_V1V2 && tb.getType() == Type.DELETE)
			if (tp.endsLine(ignored)) {
				currentBaseLine += settings.lineHeight;
				currentAdvance = 0;
			}
		}	
	}

	/**
	 * Only used for rendering
	 * @param anchor
	 */
	public void setVisibleAnchor(TextBlock anchor) {
		textRenderer.setVisibleAnchor(anchor);
	}
	
	public float getMotionPause() {
		return motionPause;
	}

	public void setMotionPause(float motionPause) {
		this.motionPause = motionPause;
	}
	
	public void updateSpeed() {
		settings.speedNormalizer.updateDocumentSpeed();
	}
	
	public float getDocumentWidth() {
		return settings.lineWidth;
	}
}