/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.diffanimation;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;

import fr.aviz.diffanimation.Document.Version;
import fr.aviz.diffanimation.TextBlock.TextWord;
import fr.aviz.diffanimation.TextBlock.Type;
import fr.aviz.diffanimation.gui.GUIUtils;
import fr.aviz.diffanimation.overview.DiffOverview;

public class SpeedNormalizer {
		
	/**
	 * Nominal animation speed (linear component).
	 * These values are set so that speed = 1 yields a "visually normal" pace
	 * for changes of reasonable magnitude, e.g, not too fast, nor too slow.
	 */
	public static final float NOMINAL_LINEAR_ANIMATION_INCREMENT = 0.007f;
	/**
	 * Nominal animation speed (quadratic component = ease out).
	 * These values are set so that speed = 1 yields a "visually normal" pace
	 * for changes of reasonable magnitude, e.g, not too fast, nor too slow.
	 */
	public static final float NOMINAL_QUADRATIC_ANIMATION_INCREMENT = 0.05f;
	/**
	 * This is the time at the beginning + at the end of the animation during which
	 * things do not move (expressed as a fraction of the total animation time).
	 * This leaves some time for highlights to fade in and fade out. 
	 * This value is for changes of reasonable magnitude. It might be recalibrated.
	 */
	public static final float NOMINAL_MOTION_PAUSE = 0.6f;
	
	public static final float MAX_SPEED_FACTOR = 3;
	public static final float MIN_SPEED_FACTOR = 0.2f;
	public static final float TURBO_FACTOR = 2f;
	
	static class WordMotion {
		Point2D.Float p1 = new Point2D.Float();
		Point2D.Float p2 = new Point2D.Float();
		float area = 0;
		
		void set(Point2D.Float p1, Point2D.Float p2, float area) {
			this.p1.setLocation(p1);
			this.p2.setLocation(p2);
			this.area = area;
		}
	}
	
//	static {
//		ExecTime.enable(false);
//	}
	
	protected DiffOverview overview;
	protected DiffAnimator detailview;
	protected DocumentAnimator document = null;
	ArrayList<WordMotion> wordMotions = new ArrayList<WordMotion>();
	int wordMotionCount = 0;
	float motionMagnitude = 0;
	
	public SpeedNormalizer(DiffOverview overview, DiffAnimator detailview) {
		this.overview = overview;
		this.detailview = detailview;
	}
	
	/**
	 * Called when new revisions have just been loaded and processed. The layout
	 * objects (paragraphs, blocks and words) have changed.
	 */
	public void documentChanged(DocumentAnimator doc) {
		this.document = doc;
		computeVisibleWordMotion();
		computeMotionMagnitude();
		updateDocumentSpeed();
	}
	
	public void setViewOfInterest(boolean validate) {
		if (validate || DiffAnimator.WORD_DISPLACEMENT_DEBUG) {
			computeVisibleWordMotion();
			computeMotionMagnitude();
			updateDocumentSpeed();
		}
	}
	
	private void clearWordMotions() {
		wordMotionCount = 0;
	}
	
	private void addWordMotion(Point2D.Float p1, Point2D.Float p2, float area) {
		WordMotion wm;
		wordMotionCount++;
		if (wordMotions.size() < wordMotionCount) {
			wm = new WordMotion();
			wordMotions.add(wm);
		} else {
			wm = wordMotions.get(wordMotionCount-1);
		}
		wm.set(p1, p2, area);
	}
		
	/**
	 * Computes the visible motion of words.
	 */
	protected void computeVisibleWordMotion() {
		
//		ExecTime.start("Word motion");

		clearWordMotions();
		if (document == null)
			return;
		
		// Y scrolling at V1 and V2
		float[] scroll = overview.getStartEndScrollingVector();
		float top1 = scroll[0];
		float top2 = scroll[1];
		
		// Visible document range
		float viewHeight = overview.barH2docH(overview.thumbHeight());
		
		// Document scale
		float scale = detailview.getScale();

		float[] geom;
		float w1, h1, w2, h2, area;
		Point2D.Float mid1 = new Point2D.Float();
		Point2D.Float mid2 = new Point2D.Float();
		float hloc; // Horizontal location within the word: 0 = left, 0.5 = middle, 1 = right
		float hloc1, hloc2;
		
		for (TextParagraph tp : document.getParagraphs()) {
			if (isVisible(top1, top2, viewHeight, tp.getPrecomputedGeometry(document.getStartVersion()), tp.getPrecomputedGeometry(document.getEndVersion()))) {
				for (TextBlock tb : tp.getBlocks()) {
					if (isVisible(top1, top2, viewHeight, tb.getPrecomputedGeometry(document.getStartVersion()), tb.getPrecomputedGeometry(document.getEndVersion()))) {
						if (tb.getType() == Type.STABLE) {
							hloc1 = 0.5f;
							hloc2 = 0.5f;
						} else {
							hloc1 = 0;
							hloc2 = 1;
						}
						for (TextWord tw : tb.getWords()) {
							if (isVisible(top1, top2, viewHeight, tw.getPrecomputedGeometry(document.getStartVersion()), tw.getPrecomputedGeometry(document.getEndVersion()))) {
								if (tw.getText().trim().length() > 0) {
									for (hloc = hloc1; hloc <= hloc2; hloc++) {
										// Get trajectory endpoints in document's coordinates
										geom = tw.getPrecomputedGeometry(document.getStartVersion());
										mid1.x = (geom[0] * (1-hloc) + geom[2] * hloc);
										mid1.y = (geom[1] * (1-hloc) + geom[3] * hloc);
										w1 = geom[1] - geom[0];
										h1 = geom[3] - geom[2];
										geom = tw.getPrecomputedMotion(document.getStartVersion(), document.getEndVersion());
										mid2.x = mid1.x + (geom[0] * (1-hloc) + geom[2] * hloc);
										mid2.y = mid1.y + (geom[1] * (1-hloc) + geom[3] * hloc);
										w2 = geom[1] - geom[0];
										h2 = geom[3] - geom[2];
										
										// Convert into view's coordinates
										mid1.x = detailview.getPageX() + mid1.x * scale;
										mid1.y = (mid1.y - top1) * scale;
										w1 *= scale;
										h1 *= scale;
										mid2.x = detailview.getPageX() + mid2.x * scale;
										mid2.y = (mid2.y - top2) * scale;
										w2 *= scale;
										h2 *= scale;
										
										// Cut the vectors so they all fit in the page
										// (skip up if they end up null-length)
										if (!cut(mid1, mid2, 0, true))
											continue;
										if (!cut(mid1, mid2, viewHeight * scale, false))
											continue;
										
										// Average word area
										area = Math.max(w1*h1, w2*h2);
										
										addWordMotion(mid1, mid2, area);
									}
								}
							}
						}
					}
				}
			}
		}
		
//		ExecTime.end();
	}
	
	/**
	 * An object is invisible if it never appears in the view during the animation.
	 */
	private static boolean isVisible(float y1, float y2, float h, float[] geom0, float[] geom1) {
		if (geom0[3] < y1 && geom1[3] < y2)
			return false;
		if (geom0[1] > y1 + h && geom1[1] > y2 + h)
			return false;
		return true;
	}
	
	private boolean cut(Point2D.Float p1, Point2D.Float p2, float y, boolean after) {
		return cutone(p1, p2, y, after) && cutone(p2, p1, y, after);
	}
	
	private boolean cutone(Point2D.Float p1, Point2D.Float p2, float y, boolean after) {
		if (after) {
			if (p1.y < y && p2.y < y)
				return false;
			if (p1.y < y) {
				if (p1.y == p2.y)
					p1.x = p2.x;
				else
					p1.x = p1.x + (y - p1.y) / (p2.y - p1.y) * (p2.x - p1.x);
				p1.y = y;
			}
		} else {
			if (p1.y > y && p2.y > y)
				return false;
			if (p1.y > y) {
				if (p1.y == p2.y)
					p1.x = p2.x;
				else
					p1.x = p1.x + (y - p1.y) / (p2.y - p1.y) * (p2.x - p1.x);
				p1.y = y;
			}
		}
		return true;
	}

	protected void paintVisibleWordMotion(Graphics2D g) {
		
		Color oldcol = g.getColor();
		Color debugColor = GUIUtils.multiplyAlpha(Color.red, 0.3f);
		g.setColor(debugColor);

		WordMotion wm;
		Point2D.Float mid1, mid2;
		for (int i=0; i<wordMotionCount; i++) {
			wm = wordMotions.get(i);
			mid1 = wm.p1;
			mid2 = wm.p2;
			
		    g.fillRect((int)mid1.x - 1, (int)mid1.y - 1, 3, 3);			
		    g.fillRect((int)mid2.x - 1, (int)mid2.y - 1, 3, 3);			
		    g.draw(new Line2D.Float(mid1, mid2));
		}
		
		g.setColor(Color.white);
		g.fillRect(10, 0, 150, 15);
		g.setColor(Color.black);
		g.drawString(getMotionMagnitude() + "", 10, 14);
		
	    g.setColor(oldcol);		
	}
	
	private static float distSq(float x1, float y1, float x2, float y2) {
		return (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1);
	}
	
	protected void computeMotionMagnitude() {
		
		/*motionMagnitude = 0;
		WordMotion wm;
		for (int i=0; i<wordMotionCount; i++) {
			wm = wordMotions.get(i);
			motionMagnitude += Math.pow(wm.p1.distance(wm.p2), 2);// * wm.area;
		}
		//
		//motionMagnitude = (float)Math.pow(motionMagnitude, 1/3f);
		//motionMagnitude = (float)Math.pow(motionMagnitude, 1/2f);
		motionMagnitude = (float)Math.log(1 + motionMagnitude);
		*/
		///////////////////////////////////////////////
		
		// TODO: Optimize
		
//		ExecTime.start("MeanDelta");
		
		float maxDeltaMotionSq = 0;
		float deltaMotionSq;
		WordMotion wm1, wm2;
		
		// Start with max absolute displacement (relative to the window)
		for (int i=0; i<wordMotionCount; i++) {
			wm1 = wordMotions.get(i);
			deltaMotionSq = distSq(wm1.p2.x - wm1.p1.x, wm1.p2.y - wm1.p1.y, 0, 0);
			if (deltaMotionSq > maxDeltaMotionSq)
				maxDeltaMotionSq = deltaMotionSq;
		}
		
		// Continue with max relative displacement
		for (int i=0; i<wordMotionCount; i++) {
			wm1 = wordMotions.get(i);
			for (int j=i+1; j<wordMotionCount; j++) {
				wm2 = wordMotions.get(j);
				deltaMotionSq = distSq(wm1.p2.x - wm1.p1.x, wm1.p2.y - wm1.p1.y, wm2.p2.x - wm2.p1.x, wm2.p2.y - wm2.p1.y);
				if (deltaMotionSq > maxDeltaMotionSq)
					maxDeltaMotionSq = deltaMotionSq;
			}
		}
		motionMagnitude = (float)Math.sqrt(maxDeltaMotionSq);
		motionMagnitude = (float)Math.log(1 + motionMagnitude);
		
//		ExecTime.end();
	}

	public float getMotionMagnitude() {
		return motionMagnitude;
	}
	
	/**
	 * 
	 */
	public void updateDocumentSpeed() {
		
		if (document == null)
			return;
		
		float speed_factor = 1;

		//
		// 1 -- Recalibrate speed according to apparent motion magnitude
		//
		if (document.settings.autoSpeed) {
			// NOTE:
			// The following equation has been estimated empirically using several values
			// for animationMagnitude and trying to keep perceived visual pace constant.
			// The speed variable should be roughly proportional to the log of the
			// animationMagnitude.
			// The coefficients come from a linear regression on several data points.
			// Maybe it's worth performing more formal measures.
			// FIXME: the fit is not good. Find a better measure for motionMagnitude.
			speed_factor = -0.46f * motionMagnitude + 3.9f;
			
			// Keep speed_factor within bounds
			if (speed_factor < MIN_SPEED_FACTOR)
				speed_factor = MIN_SPEED_FACTOR;
			else if (speed_factor > MAX_SPEED_FACTOR)
				speed_factor = MAX_SPEED_FACTOR;
		}
		
		//
		// 2 -- Compute the time needed to pause the motion for animating highlights
		//
		if (document.settings.transientHighlights)
			document.motionPause = NOMINAL_MOTION_PAUSE;
		else
			document.motionPause = 0;
		
		// Try to recalibrate the length of motion pause
		// FIXME: This is a temporary tweak
		document.motionPause = document.motionPause * speed_factor;
		if (document.motionPause < 0.05f)
			document.motionPause = 0.05f;
		if (document.motionPause > 0.8f)
			document.motionPause = 0.8f;
		//System.out.println(speed_factor + " " + document.motionPause);
		
		//
		// 3 -- Turbo-specific recalibration: Boosts an animation in case it's in a middle of a sequence.
		//
		float linear_speed_correction = 1;
		if (document.turbo) {
			speed_factor *= TURBO_FACTOR;
			// This quite complicated piece of code ensures that there is no visual jitter
			// due to the change in the value of motionPause.
			if (document.animation_coeff().get() != 0 && document.animation_coeff().get() != 1) {
				double cur = document.animation_coeff().get();
				double dest = document.animation_coeff().directGet();
				if (cur != dest) {
					float mt = document.getMotionAnimationState();
					if (cur != mt) {
						document.animation_coeff().directSet(mt);
						document.animation_coeff().set(dest);
						if (dest != mt)
							linear_speed_correction = 1 / ((float)Math.abs(dest-mt));
					}
				}
			}
			document.motionPause = 0;
		}
		
		// 
		// 4 -- Now compute linear and quadratic speed
		//
		double final_quad_speed = NOMINAL_QUADRATIC_ANIMATION_INCREMENT * (1 - document.motionPause) * document.settings.speed * speed_factor;
		double final_linear_speed = NOMINAL_LINEAR_ANIMATION_INCREMENT * (1 - document.motionPause) * document.settings.speed * speed_factor * linear_speed_correction;
		
		// 
		// 5 -- If there is no change in the document, be very fast
		//
		if (document.getChangeMagnitude() == 0) {
			final_quad_speed = 0;
			final_linear_speed = 0.5f;
		}

		
		// NOTE:
		//   Animation constants have been tuned so that speed = 1 yields a
		//   "visually normal" pace for changes of reasonable magnitude, e.g,
		//   not too fast, not too slow.
		document.animation_coeff().setIncrement(
			final_quad_speed,
			final_linear_speed
		);
//int estimate_ms = (int)(document.animation_coeff.estimateAnimationTime(1) * 1000);
//System.err.println("Animation speed updated to " + ExecTime.toString(estimate_ms) + " " + document);
	}
	
}
