/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.diffanimation;

import java.awt.geom.Rectangle2D;
import java.util.ArrayList;

import fr.aviz.diffanimation.Document.Version;

/**
 * This class corresponds to the text blocks associated to a two-versionned text.
 * 
 * Each block represents a series of character that belongs to the same status among
 * STABLE / INSERT / DELETE / MOVE
 *  
 *  Blocks are split into words. Each word is defined as a sequence of characters + the 
 *  separator character just after if it exists.
 *  
 * @author chevalie
 */
public class TextBlock {
	public enum Type {
		/** Stable text */
		STABLE,
		/** Insertion operator */
        INSERT,
        /** Deletion operator */
        DELETE,
        /** Move operator */     // Not used yet
        MOVE
    }
	
	public static final float MIN_SCALE = 0.0000001f;
	
	protected Type type;           // Type of the block {STABLE,INSERT,DELETE,MOVE}
	protected String blocktext;    // The whole text
	private int indexParagraph_v1; // Indices of the block into a paragraph
	private int indexParagraph_v2;
	private int indexDoc_v1;       // Indices of the block into the document
	private int indexDoc_v2;       
	private int length;            
	private boolean endsLine;      // if the blocks ends with '\n'
	private ArrayList<TextWord> words;
	private float scalex;
	private float scaley;
	private float [] xbounds = new float[] {0,0};
	private float [] geometry = new float[] {0,0,0,0};	
	private float [] geometry_v1 = new float[] {0,0,0,0};	
	private float [] geometry_v2 = new float[] {0,0,0,0};
	private float [] geometry_v1v2 = new float[] {0,0,0,0};
	private float [] geometry_start = new float[] {0,0,0,0};
	private float [] geometry_end = new float[] {0,0,0,0};
	private float [] motion_v1_v2 = new float[] {0,0,0,0};
	private float [] motion_v1_v1v2 = new float[] {0,0,0,0};
	private float [] motion_v1v2_v2 = new float[] {0,0,0,0};
	private float [] motion = new float[] {0,0,0,0};
	private float [] null_motion = new float[] {0,0,0,0};
	private boolean []wrap_v1_v2 = new boolean[] {false, false};
	private boolean []wrap_v1_v1v2 = new boolean[] {false, false};
	private boolean []wrap_v1v2_v2 = new boolean[] {false, false};
	private boolean []wrap = new boolean[] {false, false};
	
	ArrayList<Rectangle2D.Float> shape = new ArrayList<Rectangle2D.Float>();
	
	public TextBlock(int indexDoc, 
					 int length, 
					 String text, 
					 Type type,
					 boolean endsLine) {
		this.indexDoc_v1= indexDoc;
		this.indexDoc_v2 = indexDoc;
		this.type = type;
		this.endsLine = endsLine;
		this.blocktext = text;
		this.length = length;
		this.words = new ArrayList<TextWord>();
		this.scalex = 1.f;
		this.scaley = 1.f;
		splitIntoWords();
	}

	/**
	 * Gives the type of the block
	 * @return Type.{STABLE,DELETE,INSERT,MOVE}
	 */
	public Type getType() {
		return type;
	}
	
	/**
	 * Gives the first character index of the block in the document
	 * @param version the version (V1 / V2)
	 * @return index of the block in the document at time
	 */
	public int getIndexDocument(Version version) {
		if (version == Version.V1)
			return indexDoc_v1;
		else
			return indexDoc_v2;
	}
	
	
	/**
	 * Gives the first character index of the block in the paragraph
	 * @param version, the version (V1 / V2)
	 * @return
	 */
	public int getIndexParagraph(Version version) {
		if (version == Version.V1)
			return indexParagraph_v1;
		else
			return indexParagraph_v2;
	}
	
	/**
	 * Gives the whole number of characters of the block
	 * @return length
	 */
	public int length() {
		return length;
	}
	
	
	/**
	 * Gives the whole number of characters of the block at version
	 * @param version
	 * @return length
	 */
	public int length(Version version) {
		if (type == Type.STABLE || version == Version.V1_V2)
			return length;
		else if (type == Type.INSERT)
			return version == Version.V1 ? 0 : length;
		else
			return version == Version.V2 ? 0 : length;
	}
	
	public void setPrecomputedWrapUpLeft(Version vstart, Version vend, boolean value) {
		if (vstart == Version.V1 && vend == Version.V2)
			wrap_v1_v2[0] = value;
		else if (vstart == Version.V1 && vend == Version.V1_V2) 
			wrap_v1_v1v2[0] = value;
		else if (vstart == Version.V1_V2 && vend == Version.V2)
			wrap_v1v2_v2[0] = value;
	}
	
	
	public boolean getPrecomputedWrapUpLeft(Version vstart, Version vend) {
		if (vstart == Version.V1 && vend == Version.V2)
			return wrap_v1_v2[0];
		else if (vstart == Version.V1 && vend == Version.V1_V2) 
			return wrap_v1_v1v2[0];
		else if (vstart == Version.V1_V2 && vend == Version.V2)
			return wrap_v1v2_v2[0];
		return false;
	}
	
	public boolean getPrecomputedWrapDownRight(Version vstart, Version vend) {
		if (vstart == Version.V1 && vend == Version.V2)
			return wrap_v1_v2[1];
		else if (vstart == Version.V1 && vend == Version.V1_V2) 
			return wrap_v1_v1v2[1];
		else if (vstart == Version.V1_V2 && vend == Version.V2)
			return wrap_v1v2_v2[1];
		return false;
	}
	
	
	public void setPrecomputedWrapDownRight(Version vstart, Version vend, boolean value) {
		if (vstart == Version.V1 && vend == Version.V2)
			wrap_v1_v2[1] = value;
		else if (vstart == Version.V1 && vend == Version.V1_V2) 
			wrap_v1_v1v2[1] = value;
		else if (vstart == Version.V1_V2 && vend == Version.V2)
			wrap_v1v2_v2[1] = value;
	}
	
	
	
	
	
	
	public void setWrap(Version vstart, Version vend) {
		if (vstart == Version.V1 && vend == Version.V2)
			wrap = wrap_v1_v2;
		else if (vstart == Version.V1 && vend == Version.V1_V2) 
			wrap = wrap_v1_v1v2;
		else if (vstart == Version.V1_V2 && vend == Version.V2)
			wrap = wrap_v1v2_v2;
		else {
			if (wrap == null) {
				wrap = new boolean[]{false, false};
			}
			else {
				wrap[0] = false;
				wrap[1] = false;
			}
		}
	}
	
	public boolean getWrapUpLeft() {
		return wrap[0];
	}
	
	public boolean getWrapDownRight() {
		return wrap[1];
	}
	
	private void computeXbounds(float lineWidth, Version vstart, Version vend){
		float []geom_start = getStartGeometry();
		float []geom_end = getEndGeometry();
		
		if (type == Type.STABLE || getWrapDownRight()) {
			xbounds[0] = 0;
			xbounds[1] = lineWidth;
		}
		
		else {
			float x0 = geom_start[0]; 
			if (type == Type.DELETE && vend != Version.V1_V2)
				x0 = geom_end[0];
			float width = lineWidth * scalex;
		
			float xmin = x0*(1-scalex);
			float xmax = xmin+width;
			xbounds[0] = xmin;
			xbounds[1] = xmax;
		}
	}
	
	public ArrayList<Rectangle2D.Float> getShape() {
		return shape;
	}
	
	
	public void computeShape(Settings settings, Version vstart, Version vend) {
		float lineWidth = settings.lineWidth;
		float lineHeight = settings.lineHeight;
		float zoomedLineHeight = lineHeight * scaley;

		float x1 = geometry[0];
		float y1 = geometry[1];
		float x2 = geometry[2];
		float y2 = geometry[3];
		
		float height = y2 - y1;
		float width = x2 - x1;
		
		computeXbounds(lineWidth, vstart, vend);
		shape.clear();
		
		// -- 1: empty block case
		if (height == 0 && width == 0) {
			Rectangle2D.Float r = new Rectangle2D.Float(x1, y1-zoomedLineHeight/2.f, 0, zoomedLineHeight);
			shape.add(r);
			return;
		}
		
		// -- 2: non-rescaled block
		if (scalex == 1 && scaley == 1) {
			while (x1 < 0) {
				x1 += lineWidth;
				y1 -= lineHeight;	
			}
			while (x1 > lineWidth) {
				x1 -= lineWidth;
				y1 += lineHeight;
			}
			while (x2 < 0) {
				x2 += lineWidth;
				y2 -= lineHeight;	
			}
			while (x2 > lineWidth) {
				x2 -= lineWidth;
				y2 += lineHeight;
			}
			
			height = y2 - y1;
			
			// 1: One-line
			if (height <= lineHeight+0.1f) {
				Rectangle2D.Float r = new Rectangle2D.Float(x1, y1, x2-x1, lineHeight);
				shape.add(r);
				return;
			}
			
			else {
				// Top line
				Rectangle2D.Float r1 = new Rectangle2D.Float(x1, y1, lineWidth-x1, lineHeight);
				shape.add(r1);
			
				// Bottom line 
				Rectangle2D.Float r2 = new Rectangle2D.Float(0, y2-lineHeight, x2, lineHeight);
				shape.add(r2);

				// Full midlines
				Rectangle2D.Float r3 = new Rectangle2D.Float(0, y1+lineHeight, lineWidth,y2-y1-2*lineHeight);
				if (r3.height > 1f)
					shape.add(r3);
			
				return;
			}
		}
	
		
		// -- 3: rescaled case
		
		// -- 3.1: case not extended
		if (getPrecomputedWrapDownRight(vstart, vend)) {
			while (x1 < 0) {
				x1 += lineWidth;
				y1 -= lineHeight;	
			}
			while (x1 > lineWidth) {
				x1 -= lineWidth;
				y1 += lineHeight;
			}
			while (x2 < 0) {
				x2 += lineWidth;
				y2 -= lineHeight;	
			}
			while (x2 > lineWidth) {
				x2 -= lineWidth;
				y2 += lineHeight;
			}
			
			height = y2 - y1;
			
			// 1: One-line
			if (height <= zoomedLineHeight+0.1f) {
				Rectangle2D.Float r = new Rectangle2D.Float(x1, y1, x2-x1, zoomedLineHeight);
				shape.add(r);
				return;
			}
			
			else {
				// Top line
				Rectangle2D.Float r1 = new Rectangle2D.Float(x1, y1, lineWidth-x1, zoomedLineHeight);
				shape.add(r1);
			
				// Bottom line 
				Rectangle2D.Float r2 = new Rectangle2D.Float(0, y2-zoomedLineHeight, x2, zoomedLineHeight);
				shape.add(r2);

				// Full midlines
				Rectangle2D.Float r3 = new Rectangle2D.Float(0, y1+zoomedLineHeight, lineWidth,y2-y1-2*zoomedLineHeight);
				if (r3.height >= lineHeight)
					shape.add(r3);
			
				return;
			}
		}
		
		// -- 3.2: case extended / shrinked
		else {
			while (x1 < 0) {
				x1 += lineWidth;
				y1 -= zoomedLineHeight;	
			}
			while (x1 > lineWidth) {
				x1 -= lineWidth;
				y1 += zoomedLineHeight;
			}
			while (x2 < 0) {
				x2 += lineWidth;
				y2 -= zoomedLineHeight;	
			}
			while (x2 > lineWidth) {
				x2 -= lineWidth;
				y2 += zoomedLineHeight;
			}
			
			height = y2 - y1;
			
			// 1: One-line
			if (height <= zoomedLineHeight+0.1f) {
				Rectangle2D.Float r = new Rectangle2D.Float(x1, y1, x2-x1, zoomedLineHeight);
				shape.add(r);
				return;
			}
			
			else {
				// Top line
				Rectangle2D.Float r1 = new Rectangle2D.Float(x1, y1, (xbounds[1]-x1), zoomedLineHeight);
				shape.add(r1);
			
				// Bottom line 
				Rectangle2D.Float r2 = new Rectangle2D.Float(xbounds[0], y2-zoomedLineHeight, x2-xbounds[0], zoomedLineHeight);
				shape.add(r2);

				// Full midlines
				Rectangle2D.Float r3 = new Rectangle2D.Float(xbounds[0], y1+zoomedLineHeight, xbounds[1]-xbounds[0],y2-y1-2*zoomedLineHeight);
				if (r3.height >= zoomedLineHeight-0.1f)
				shape.add(r3);
				return;
			}
		}
	}
	
	
	public void setGeometry(float x1, float y1, float x2, float y2) {
		geometry[0] = x1;
		geometry[1] = y1;
		geometry[2] = x2;
		geometry[3] = y2;
	}
	
	public float[] getGeometry(){
		return geometry;
	}
	
	public void setStartGeometry(Version version) {
		if (version == Version.V1)
			geometry_start = geometry_v1;
		else if (version == Version.V2)
			geometry_start = geometry_v2;
		else if (version == Version.V1_V2)
			geometry_start = geometry_v1v2;
	}
	
	public float[] getStartGeometry() {
		return geometry_start;
	}
	
	public void setEndGeometry(Version version) {
		if (version == Version.V1)
			geometry_end = geometry_v1;
		else if (version == Version.V2)
			geometry_end = geometry_v2;
		else if (version == Version.V1_V2)
			geometry_end = geometry_v1v2;
	}
	
	public float[] getEndGeometry() {
		return geometry_end;
	}
	
	public void setPrecomputedGeometry(float x1, float y1, float x2, float y2, Version version) {
		if (version == Version.V1) {
			geometry_v1[0] = x1;
			geometry_v1[1] = y1;
			geometry_v1[2] = x2;
			geometry_v1[3] = y2;
		} else if (version == Version.V2) {
			geometry_v2[0] = x1;
			geometry_v2[1] = y1;
			geometry_v2[2] = x2;
			geometry_v2[3] = y2;
		} else {
			geometry_v1v2[0] = x1;
			geometry_v1v2[1] = y1;
			geometry_v1v2[2] = x2;
			geometry_v1v2[3] = y2;
		}
	}
	
	/**
	 * Necessary after word justification.
	 * @param version
	 * @return
	 */
	public void correctPrecomputedGeometry(Version version) {
		if (words.size() == 0)
			return;
		TextWord firstword = words.get(0);
		TextWord lastword = words.get(words.size() - 1);
		if (version == Version.V1) {
			geometry_v1[0] = firstword.geometry_v1[0];
			geometry_v1[1] = firstword.geometry_v1[1];
			geometry_v1[2] = lastword.geometry_v1[2];
			geometry_v1[3] = lastword.geometry_v1[3];
		} else if (version == Version.V2) {
			geometry_v2[0] = firstword.geometry_v2[0];
			geometry_v2[1] = firstword.geometry_v2[1];
			geometry_v2[2] = lastword.geometry_v2[2];
			geometry_v2[3] = lastword.geometry_v2[3];
		} else {
			geometry_v1v2[0] = firstword.geometry_v1v2[0];
			geometry_v1v2[1] = firstword.geometry_v1v2[1];
			geometry_v1v2[2] = lastword.geometry_v1v2[2];
			geometry_v1v2[3] = lastword.geometry_v1v2[3];
		}
	}
	
	public float[] getPrecomputedGeometry(Version version) {
		if (version == Version.V1)
			return geometry_v1;
		else if (version == Version.V2)
			return geometry_v2;
		else 
			return geometry_v1v2;
	}
	
	public void setPrecomputedMotion(Version vstart, Version vend, float m_x1, float m_y1, float m_x2, float m_y2) {
		if (vstart == Version.V1 && vend == Version.V2) {
			motion_v1_v2[0] = m_x1;
			motion_v1_v2[1] = m_y1;
			motion_v1_v2[2] = m_x2;
			motion_v1_v2[3] = m_y2;
		} 
		else if  (vstart == Version.V1 && vend == Version.V1_V2){
			motion_v1_v1v2[0] = m_x1;
			motion_v1_v1v2[1] = m_y1;
			motion_v1_v1v2[2] = m_x2;
			motion_v1_v1v2[3] = m_y2;
		} 
		else if  (vstart == Version.V1_V2 && vend == Version.V2){
			motion_v1v2_v2[0] = m_x1;
			motion_v1v2_v2[1] = m_y1;
			motion_v1v2_v2[2] = m_x2;
			motion_v1v2_v2[3] = m_y2;
		} 
	}
	
	public float[] getPrecomputedMotion(Version vstart, Version vend){
		if (vstart == vend)
			return null_motion;
		else if (vstart == Version.V1 && vend == Version.V2)
			return motion_v1_v2;
		else if (vstart == Version.V1 && vend == Version.V1_V2) 
			return motion_v1_v1v2;
		else if (vstart == Version.V1_V2 && vend == Version.V2)
			return motion_v1v2_v2;
		return null;
	}
	
	public void setMotion(Version vstart, Version vend) {
		if (vstart == vend)
			motion = null_motion;
		else if (vstart == Version.V1 && vend == Version.V2)
			motion = motion_v1_v2;
		else if (vstart == Version.V1 && vend == Version.V1_V2)
			motion = motion_v1_v1v2;
		else if (vstart == Version.V1_V2 && vend == Version.V2)
			motion = motion_v1v2_v2;
	}

	public float[] getMotion() {
		return motion;
	}
	
	
	
	public float getScalex() {
		return this.scalex;
	}
	
	public float getScaleY() {
		return this.scaley;
	}
	
	public void setScales(float scalex, float scaley) {
		this.scalex = scalex;
		this.scaley = scaley;
	}
		
	public void computeScales(float motion_t, Version vstart, Version vend) {
		scalex = 1f;
		scaley = 1f;
	
		if (vend == vstart)
			return;
		
		Type stable = null;
		if (vend == Version.V1_V2) 
			stable = Type.DELETE;
		else if (vstart == Version.V1_V2) 
			stable = Type.INSERT;
		
		if (type == Type.STABLE || type == stable) 
			return;
		
		
		if (type == Type.DELETE) {
			scalex = 1f - motion_t;
			scaley = 1f - motion_t;
		}
		else if (type == Type.INSERT) {
			scalex = motion_t;
			scaley = motion_t;
		}
	
		
		// AffineTransform needs non-null factors
		if (scalex == 0)
			scalex = MIN_SCALE;
		if (scaley == 0)
			scaley = MIN_SCALE;
	}
	
	
	/**
	 * Indicates if the block last character is a '\n'
	 * @return true if the block ends with a '\n'
	 */
	public boolean endsLine() {
		return endsLine;
	}
	
	
	/**
	 * Set document index of the first character of the block in the 
	 * document
	 * @param version the version (V1 / V2)
	 * @param index the new index value
	 */
	public void setIndexDocument(Version version, int index) {
		if (version == Version.V1)
			this.indexDoc_v1 = index;
		else 
			this.indexDoc_v2 = index;
	}
	
	
	/**
	 * Set document index of the first character of the block in the 
	 * document
	 * @param version the version (V1 / V2)
	 * @param index the new index value
	 */
	public void setIndexParagraph(Version version, int index) {
		if (version == Version.V1)
			this.indexParagraph_v1 = index;
		else 
			this.indexParagraph_v2 = index;
	}
	
	
	/**
	 * Sets the type of the block
	 * @param type the new type (STABLE,INSERT,MOVE,DELETE)
	 */
	public void setType(Type type) {
		this.type = type;
	}
	
	
	
	/**
	 * Gives the string contained in the block
	 * @return text of the block
	 */
	public String getText() {
		return blocktext;
	}

	
	/**
	 * Gives the list of words in the block 
	 * A word contains the separator at its ends
	 * "examples ", "of ", "words."  
	 * @return list of words
	 */
	public ArrayList<TextWord> getWords() {
		return words;
	}
	

	/**
	 * Positions comparison with tb  (-1 : before, 0: same position, 1: after)
	 */
	public int compareTo(TextBlock tb) {
		if (indexDoc_v1 == tb.getIndexDocument(Version.V1)) {
			return tb.getType() == Type.DELETE ? -1 : 1;
		}
		
		return indexDoc_v1 < tb.getIndexDocument(Version.V1) ? -1 : 1;
	}
	
	
	
	/**
	 * Splits the block into words
	 * "Example: Example of a sentence split." 
	 *  ->  "Example: ", "Example ", "of ", "a ", "sentence ", "split."
	 */
	public void splitIntoWords() {
		words.clear();
		String remainingtext = this.blocktext;
		int nextIndex = getNextSeparatorIndex(remainingtext);
		int wordIndex = 0;
		
		while (nextIndex != -1 && wordIndex + nextIndex < blocktext.length()-1) {
			String word = remainingtext.substring(0, nextIndex+1);  // we put the separator in the word
			TextWord tw = new TextWord(word, wordIndex);
			words.add(tw);
			wordIndex += word.length();
			remainingtext = remainingtext.substring(nextIndex+1);
			nextIndex = getNextSeparatorIndex(remainingtext);
		}
		
		// Last word
		TextWord tw = new TextWord(remainingtext, wordIndex);
		words.add(tw);
	}

	
	
	/**
	 * Gives the index of the first separator in a string
	 * 
	 * @param text 
	 * @return index of the first separator, -1 if there are none
	 */
	private int getNextSeparatorIndex(String text) {
		int nextSpace = text.indexOf(' ');
		int nextTab = text.indexOf('\t');
		int nextEnter = text.indexOf('\n');
		
		int index;
		if (nextSpace == -1 && nextEnter == -1 && nextTab == -1)
			return -1;
		
		
		if (nextSpace != -1) {
			if (nextTab != -1) index = nextSpace < nextTab ? nextSpace : nextTab;
			else index = nextSpace;
			if (nextEnter != -1) index = index < nextEnter ? index : nextEnter;
			return index;
		}
		
		else {
			if (nextTab != -1) {
				if (nextEnter != -1) return nextTab < nextEnter ? nextTab : nextEnter;
				else return nextTab;
			}
			else return nextEnter;
		}
	}
	
	

	
	
	public void computeWordBounds(Settings settings) {
		float ascent = settings.nominalTextLayout.getAscent() + settings.nominalTextLayout.getLeading();
		float currentAdvance = this.geometry_v1[0];
		float currentBaseLine = this.geometry_v1[1]+ascent;
		
		for (TextWord tw : words) {
			String word = tw.getText();
			if (word.length() != 0) {
				float wrappingWidth = settings.lineWidth - currentAdvance;
				
				if (settings.nominalTextLayout.getAdvance() > wrappingWidth) {
					currentBaseLine += settings.lineHeight;
					currentAdvance = 0;
				}
				tw.setPrecomputedGeometry(currentAdvance, currentBaseLine, currentAdvance+settings.lineHeight, currentAdvance + settings.nominalTextLayout.getAdvance(), Version.V1);
				currentAdvance += settings.nominalTextLayout.getAdvance();
			}	
		} 
	}
	
	/**
	 * Word in a block object
	 */
	public class TextWord {
		private String text;
		private int index; // index of the first character in the block
		private float []geometry;
		private float []geometry_v1;
		private float []geometry_v2;
		private float []geometry_v1v2;
		private float []geometry_start;
		private float []geometry_end;
		private float [] motion_v1_v2;
		private float [] motion_v1_v1v2;
		private float [] motion_v1v2_v2;
		private float [] motion;
		private float advance, visibleadvance, full_word_advance_v1, full_word_advance_v2, full_word_advance_v1v2;
		
		public TextWord(String text, int index){
			this.text = text;
			this.index = index;
			this.geometry = new float[]{0,0,0,0};
			this.geometry_v1 = new float[]{0,0,0,0};
			this.geometry_v2 = new float[]{0,0,0,0};
			this.geometry_v1v2 = new float[]{0,0,0,0};
			this.motion_v1_v2 = new float[]{0,0,0,0};
			this.motion_v1_v1v2 = new float[]{0,0,0,0};
			this.motion_v1v2_v2 = new float[]{0,0,0,0};
		}
		
		public String getText() {
			return this.text;
		}
		
		public int getTrailingSpaces() {
			// words have at most one unique space
			return this.text.endsWith(" ") ? 1 : 0;
		}
	
		public int getIndex() {
			return this.index;
		}
		
		public int getSize() {
			return text.length();
		}
	
		/**
		 * The word's width, trailing whitespace included (if any).
		 * @param advance
		 */
		public float getAdvance() {
			return advance;
		}
		
		public float getVisibleAdvance() {
			return visibleadvance;
		}
		
		/**
		 * The full word's with (i.e., across blocks), trailing whitespace not included.
		 * @param advancespace
		 */
		public float getFullWordAdvance(Version version) {
			if (version == Version.V1)
				return full_word_advance_v1;
			else if (version == Version.V2)
				return full_word_advance_v2;
			else
				return full_word_advance_v1v2;
		}
		
		public void setAdvance(float advance) {
			this.advance = advance;
		}
		
		public void setVisibleAdvance(float advance) {
			this.visibleadvance = advance;
		}
		
		public void setFullWordAdvance(float advance, Version version) {
			if (version == Version.V1)
				this.full_word_advance_v1 = advance;
			else if (version == Version.V2)
				this.full_word_advance_v2 = advance;
			else 
				this.full_word_advance_v1v2 = advance;
		}
		
		public void translatex(float dx, Version version) {
			if (version == Version.V1) {
				this.geometry_v1[0] += dx;
				this.geometry_v1[2] += dx;
			} else if (version == Version.V2) {
				this.geometry_v2[0] += dx;
				this.geometry_v2[2] += dx;
			} else {
				this.geometry_v1v2[0] += dx;
				this.geometry_v1v2[2] += dx;
			}
		}
		
		public void setGeometry(float x1, float y1, float x2, float y2) {
			this.geometry[0] = x1;
			this.geometry[1] = y1;
			this.geometry[2] = x2;
			this.geometry[3] = y2;
		}
				
		public float[] getGeometry() {
			return this.geometry;
		}
		
		public void setStartGeometry(Version version) {
			if (version == Version.V1)
				this.geometry_start = this.geometry_v1;
			else if (version == Version.V2)
				this.geometry_start = this.geometry_v2;
			else if (version == Version.V1_V2)
				this.geometry_start = this.geometry_v1v2;
		}
		
		public float[] getStartGeometry() {
			return this.geometry_start;
		}
		
		public void setEndGeometry(Version version) {
			if (version == Version.V1)
				this.geometry_end = this.geometry_v1;
			else if (version == Version.V2)
				this.geometry_end = this.geometry_v2;
			else if (version == Version.V1_V2)
				this.geometry_end = this.geometry_v1v2;
		}
		
		public float[] getEndGeometry() {
			return this.geometry_end;
		}
		
		public void setPrecomputedGeometry(float x1, float y1, float x2, float y2, Version version) {
			if (version == Version.V1) {
				this.geometry_v1[0] = x1;
				this.geometry_v1[1] = y1;
				this.geometry_v1[2] = x2;
				this.geometry_v1[3] = y2;
			}
			else if (version == Version.V2) {
				this.geometry_v2[0] = x1;
				this.geometry_v2[1] = y1;
				this.geometry_v2[2] = x2;
				this.geometry_v2[3] = y2;
			}
			else {
				this.geometry_v1v2[0] = x1;
				this.geometry_v1v2[1] = y1;
				this.geometry_v1v2[2] = x2;
				this.geometry_v1v2[3] = y2;
			}
		}
		
		
		public float[] getPrecomputedGeometry(Version version) {
			if (version == Version.V1)
				return this.geometry_v1;
			else if (version == Version.V2)
				return this.geometry_v2;
			else
				return this.geometry_v1v2;
		}	
		
		
		public void setPrecomputedMotion(Version vstart, Version vend, float m_x1, float m_y1, float m_x2, float m_y2) {
			if (vstart == Version.V1 && vend == Version.V2) {
				this.motion_v1_v2[0] = m_x1;
				this.motion_v1_v2[1] = m_y1;
				this.motion_v1_v2[2] = m_x2;
				this.motion_v1_v2[3] = m_y2;
			} 
			else if  (vstart == Version.V1 && vend == Version.V1_V2){
				this.motion_v1_v1v2[0] = m_x1;
				this.motion_v1_v1v2[1] = m_y1;
				this.motion_v1_v1v2[2] = m_x2;
				this.motion_v1_v1v2[3] = m_y2;
			} 
			else if  (vstart == Version.V1_V2 && vend == Version.V2){
				this.motion_v1v2_v2[0] = m_x1;
				this.motion_v1v2_v2[1] = m_y1;
				this.motion_v1v2_v2[2] = m_x2;
				this.motion_v1v2_v2[3] = m_y2;
			} 
		}
		
		public float[] getPrecomputedMotion(Version vstart, Version vend){
			if (vstart == Version.V1 && vend == Version.V2)
				return this.motion_v1_v2;
			else if (vstart == Version.V1 && vend == Version.V1_V2) 
				return this.motion_v1_v1v2;
			else if (vstart == Version.V1_V2 && vend == Version.V2)
				return this.motion_v1v2_v2;
			return null;
		}
		
		public void setMotion(Version vstart, Version vend) {
			if (vstart == Version.V1 && vend == Version.V2)
				this.motion = this.motion_v1_v2;
			else if (vstart == Version.V1 && vend == Version.V1_V2)
				this.motion = this.motion_v1_v1v2;
			else if (vstart == Version.V1_V2 && vend == Version.V2)
				this.motion = this.motion_v1v2_v2;
		}

		public float[] getMotion() {
			return this.motion;
		}
	}


	
}