/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.diffanimation;

import java.util.ArrayList;

import fr.aviz.diffanimation.Document.Version;
import fr.aviz.diffanimation.TextBlock.Type;

/**
 * This class corresponds to the paragraphs in a consensus version of the document
 * A paragraph consists in a series of text blocks such that the last block 
 * and only the last block ends with '\n'
 * 
 * If the last block is inserted text, the paragraph structure exists only at V2
 *    (the paragraph after does not start at the beginning of the line at V1)
 * If the last block is deleted text, the paragraph structure exists only at V1
 * 
 * @author fanny
 */
public class TextParagraph {
	private ArrayList<TextBlock> blocks;

	private int index_v1;    // Index of the first character in the document
	private int index_v2;
	private int length_v1;   // number of characters
	private int length_v2;
	private int length_v1_v2;
	private float[] geometry = new float[]{0,0,0,0};
	private float[] geometry_v1 = new float[]{0,0,0,0};
	private float[] geometry_v2 = new float[]{0,0,0,0};
	private float[] geometry_v1v2 = new float[]{0,0,0,0};
	private float[] geometry_start= new float[]{0,0,0,0};
	private float[] geometry_end = new float[]{0,0,0,0};
	private float[] motion_v1_v2 = new float[]{0,0,0,0};
	private float[] motion_v1_v1v2 = new float[]{0,0,0,0};
	private float[] motion_v1v2_v2 = new float[]{0,0,0,0};
	private float[] motion = new float[]{0,0,0,0};
	private float[] null_motion = new float[]{0,0,0,0};
	
	public TextParagraph(int index) {
		this.index_v1 = index;
		blocks = new ArrayList<TextBlock>();
	}
	
	public void setGeometry(float x1, float y1, float x2, float y2) {
		geometry[0] = x1;
		geometry[1] = y1;
		geometry[2] = x2;
		geometry[3] = y2;
	}
	
	public float[] getGeometry() {
		return geometry;
	}
	
	public void setPrecomputedGeometry(float x1, float y1, float x2, float y2, Version version) {
		if (version == Version.V1) {
			geometry_v1[0] = x1;
			geometry_v1[1] = y1;
			geometry_v1[2] = x2;
			geometry_v1[3] = y2;
		}
		else if (version == Version.V2) {
			geometry_v2[0] = x1;
			geometry_v2[1] = y1;
			geometry_v2[2] = x2;
			geometry_v2[3] = y2;
		}
		else {
			geometry_v1v2[0] = x1;
			geometry_v1v2[1] = y1;
			geometry_v1v2[2] = x2;
			geometry_v1v2[3] = y2;
		}
	}
	
	
	public float[] getPrecomputedGeometry(Version version) {
		if (version == Version.V1)
			return geometry_v1;
		else if (version == Version.V2)
			return geometry_v2;
		else
			return geometry_v1v2;
	}
	
	public void setStartGeometry(Version version) {
		if (version == Version.V1)
			geometry_start = geometry_v1;
		else if (version == Version.V2)
			geometry_start = geometry_v2;
		else if (version == Version.V1_V2)
			geometry_start = geometry_v1v2;
	}
	
	public float[] getStartGeometry() {
		return geometry_start;
	}
	
	public void setEndGeometry(Version version) {
		if (version == Version.V1)
			geometry_end = geometry_v1;
		else if (version == Version.V2)
			geometry_end = geometry_v2;
		else if (version == Version.V1_V2)
			geometry_end = geometry_v1v2;
	}
	
	public float[] getEndGeometry() {
		return geometry_end;
	}
	
	public void setPrecomputedMotion(Version vstart, Version vend, float m_x1, float m_y1, float m_x2, float m_y2) {
		if (vstart == Version.V1 && vend == Version.V2) {
			motion_v1_v2[0] = m_x1;
			motion_v1_v2[1] = m_y1;
			motion_v1_v2[2] = m_x2;
			motion_v1_v2[3] = m_y2;
		} 
		else if  (vstart == Version.V1 && vend == Version.V1_V2){
			motion_v1_v1v2[0] = m_x1;
			motion_v1_v1v2[1] = m_y1;
			motion_v1_v1v2[2] = m_x2;
			motion_v1_v1v2[3] = m_y2;
		} 
		else if  (vstart == Version.V1_V2 && vend == Version.V2){
			motion_v1v2_v2[0] = m_x1;
			motion_v1v2_v2[1] = m_y1;
			motion_v1v2_v2[2] = m_x2;
			motion_v1v2_v2[3] = m_y2;
		} 
	}
	
	public float[] getPrecomputedMotion(Version vstart, Version vend){
		if (vstart == vend)
			return null_motion;
		else if (vstart == Version.V1 && vend == Version.V2)
			return motion_v1_v2;
		else if (vstart == Version.V1 && vend == Version.V1_V2) 
			return motion_v1_v1v2;
		else if (vstart == Version.V1_V2 && vend == Version.V2)
			return motion_v1v2_v2;
		return null;
	}
	
	public void setMotion(Version vstart, Version vend) {
		if (vstart == Version.V1 && vend == Version.V2)
			motion = motion_v1_v2;
		else if (vstart == Version.V1 && vend == Version.V1_V2)
			motion = motion_v1_v1v2;
		else if (vstart == Version.V1_V2 && vend == Version.V2)
			motion = motion_v1v2_v2;
	}

	public float[] getMotion() {
		return motion;
	}
	
	/**
	 * Returns the number of characters of the paragraph
	 * @param version
	 * @return length
	 */
	public int length(Version version) {
		if (version == Version.V1)
			return length_v1;
		else if (version == Version.V2)
			return length_v2;
		else 
			return length_v1_v2;
	}
	
	
	/**
	 * Returns the index of the paragraph in the document
	 * @param version the version
	 * @return paragraph index in the document stream 
	 */
	public int getIndex(Version version) {
		if (version == Version.V1)
			return index_v1;
		else
			return index_v2;
	}
	
	
	/**
	 * Returns the list of text blocks contained in the paragraph
	 * @return blocks of text
	 */
	public ArrayList<TextBlock> getBlocks() {
		return blocks;
	}
	
	
	
	/**
	 * Adds a text block in the paragraph
	 * @param tb the text block to add
	 */
	public void addTextBlock(TextBlock tb) {
		blocks.add(tb);
		if (tb.getType() != Type.INSERT)
			length_v1 += tb.length();
		if (tb.getType() != Type.DELETE)
			length_v2 += tb.length();
		length_v1_v2 += tb.length();
	}
	
	

	
	
	
	/**
	 * Returns true if the paragraph's last character is '\n'
	 * @param version the version
	 * @return true if the last character is '\n'
	 */
	/*public boolean endsLine(Version version) {
		TextBlock last = blocks.get(blocks.size()-1);
		if (version == Version.V1) {
			if (last.type == Type.INSERT)
				return false;
		}
		else  {
			if (last.type == Type.DELETE)
				return false;
		}
		return last.endsLine(); 
	}*/
	
	public boolean endsLine(Type ignored) {
		TextBlock last = blocks.get(blocks.size()-1);
		if (last.type == ignored)
			return false;
		return last.endsLine(); 
	}
	
	/**
	 * Returns true if the paragraph's last character is '\n'
	 * @param animversions the versions we consider: V1_V2, V1_V1V2 or V1V2_V2
	 * @param version the version
	 * @return true if the last character is '\n'
	 */
	/*public boolean endsLine(AnimVersions animversions, Version version) {
		TextBlock last = blocks.get(blocks.size()-1);
		if (version == Version.V1 && animversions != AnimVersions.V1V2_V2) {
			if (last.type == Type.INSERT)
				return false;
		}
		else  {
			if (last.type == Type.DELETE && animversions != AnimVersions.V1_V1V2)
				return false;
		}
		return last.endsLine(); 
	}*/
	
	
	/**
	 * @return true if the paragraph contains only inserted text
	 */
	public boolean isInsertion() {
		for (TextBlock tb : blocks) {
			if (tb.type != Type.INSERT)
				return false;
		}
		return true;
	}
	
	
	/**
	 * @return true if the paragraph contains only deleted text
	 */
	public boolean isDeletion() {
		for (TextBlock tb : blocks) {
			if (tb.type != Type.DELETE)
				return false;
		}
		return true;
	}
	
	


	/**
	 * Computes the paragraph length at V2
	 */
	public void computeLength_v2() {
		length_v2 = 0;
		for (TextBlock block : blocks) 
			if (block.getType() != Type.DELETE)
				length_v2 += block.length();
	}
}
