/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.diffanimation.overview;

import java.awt.AWTException;
import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.AffineTransform;
import java.awt.Robot;


import javax.swing.JComponent;
import javax.swing.SwingUtilities;

import fr.aviz.animation.AnimatedFloat;
import fr.aviz.animation.AnimationListener;
import fr.aviz.animation.AnimationTimer;
import fr.aviz.diffanimation.DiffAnimator;
import fr.aviz.diffanimation.DocumentAnimator;
import fr.aviz.diffanimation.DocumentAnimator.PaintMode;
import fr.aviz.utils.DocumentFactory;
import fr.aviz.utils.ExecTime;

/**
 * Overview + scrollbar
 * 
 * FIXME: There is an offset when animations are canceled
 * 
 * @author dragice
 *
 */
public class DiffOverview extends JComponent implements AnimationListener, MouseListener, MouseMotionListener, MouseWheelListener {

	// The component's width
	public static final int WIDTH = 24;
	
	// The component's vertical and horizontal internal margins
	public static final int PADDING = 2;
	
	//
	public static final int TIMELINE_DRAG_THRESHOLD = 24;
	
	// Colors
	public static final Color BORDER_COLOR = new Color(0.85f, 0.85f, 0.85f, 1f);
	public static final Color THUMB_COLOR = new Color(0.2f, 0.2f, 1f, 0.85f);
	public static final Color THUMB_FILL_COLOR = new Color(0.75f, 0.75f, 1, 0.3f);
	public static Stroke THIN_STROKE = new BasicStroke(1);
	
	public final boolean allowThumbToBeOutside = false;
	
	DiffAnimator detailView;
	ScrollingStabilizer stabilizer;
	float thumbY = 0;
	
	// -- Interaction state
	int lastmousex, lastmousey;
	boolean dragging_ = false;
	boolean draggingX = false;
	boolean draggingY = false;
	
	// -- For dragX
	Robot robot;
	int constrainedY;
	
	// Total document height in pixels, in the original document's coordinates system.
	float maxDocumentHeight = 0;
	
	AnimatedFloat opacity = new AnimatedFloat(0, AnimatedFloat.DEFAULT_QUADRATIC_INCREMENT/2, AnimatedFloat.DEFAULT_LINEAR_INCREMENT/2);

//	String[] currentFiles;
	
	/**
	 * The important thing here is the call to AnimationTimer.addListener().
	 */
	public DiffOverview(DiffAnimator detailView) {
		super();
		this.detailView = detailView;
		this.stabilizer = new ScrollingStabilizer(this, detailView);
		stabilizer.setTopBottomAnchorEnabled(!allowThumbToBeOutside);
		AnimationTimer.addListener(this);
		addMouseListener(this);
		addMouseMotionListener(this);
		addMouseWheelListener(this);
		setPreferredSize(new Dimension(WIDTH, 800));
		setFocusable(true);
		
		// Initialize robot for horizontal-constrained movement
		try {
			robot = new java.awt.Robot();
			
		} catch (AWTException e) {
			System.err.println("DiffOverview failed to init Robot");
		}
	}
	
	/**
	 * This will tell the component to repaint itself each time an animated float
	 * changes. See the call to AnimationTimer.addListener in the constructor.
	 */
	public void animateEvent() {
		stabilize();
	}
	
	public void stabilize() {
		stabilizer.stabilize();
		repaint();
	}
	
	/**
	 * In screen coordinates
	 * @return
	 */
	public float barYmin() {
		return PADDING - 1;
	}
	
	/**
	 * In screen coordinates
	 * @return
	 */
	public float barYmax() {
		return getHeight() - (PADDING - 1);
	}
	
	/**
	 * In screen coordinates
	 * @return
	 */
	public float barHeight() {
		return barYmax() - barYmin();
	}
	
	/**
	 * In screen coordinates
	 * @return
	 */
	public float thumbY() {
		return thumbY;
	}
	
	/**
	 * In screen coordinates
	 * @return
	 */
	public float thumbHeight() {
		return viewH2barH(detailView.getHeight());
	}

	/**
	 * In document coordinates
	 * @return
	 */
	public float docYmin() {
		return 0;
	}
	
	/**
	 * In document coordinates
	 * @return
	 */
	public float docYmax() {
		return maxDocumentHeight;
	}
	
	/**
	 * In document coordinates
	 * @return
	 */
	public float docHeight() {
		return docYmax() - docYmin();
	}
	
	/**
	 * In screen coordinates
	 * @return
	 */
	public float viewYmin() {
		return -bar2doc(thumbY()) * detailView.getScale();
	}
	
	/**
	 * In screen coordinates
	 * @return
	 */
	public float viewYmax() {
		return viewYmin() + viewHeight();
	}

	/**
	 * In screen coordinates
	 * @return
	 */
	public float viewHeight() {
		return docHeight() * detailView.getScale();
	}
	
	/**
	 * Converts y-coordinates.
	 * @param barY
	 * @return
	 */
	public float bar2doc(float barY) {
		return (barY - barYmin()) / barHeight() * docHeight() + docYmin(); 
	}
	
	/**
	 * Converts heights.
	 * @param barY
	 * @return
	 */
	public float barH2docH(float barH) {
		return barH / barHeight() * docHeight(); 
	}
	
	/**
	 * Converts y-coordinates.
	 * @param docY
	 * @return
	 */
	public float doc2bar(float docY) {
		return (docY - docYmin()) / docHeight() * barHeight() + barYmin(); 
	}
	
	/**
	 * Converts heights.
	 */
	public float docH2barH(float docH) {
		return docH / docHeight() * barHeight(); 
	}
	
	/**
	 * Converts y-coordinates.
	 * @param barY
	 * @return
	 */
	public float view2bar(float viewY) {
		return (viewY - viewYmin()) / viewHeight() * barHeight() + barYmin(); 
	}
	
	/**
	 * Converts y-coordinates.
	 * @param docY
	 * @return
	 */
	public float bar2view(float barY) {
		return (barY - barYmin()) / barHeight() * viewHeight() + viewYmin(); 
	}
	
	/**
	 * Converts heights.
	 */
	public float viewH2barH(float docH) {
		return docH / viewHeight() * barHeight(); 
	}
	
	public boolean isActive() {
		return detailView.getDocument() != null && viewHeight() > barHeight();
	}
	
	public void directUpdateOpacity() {
		if (isActive() && opacity.get() != 1)
			opacity.directSet(1);
		if (!isActive() && opacity.get() != 0)
			opacity.directSet(0);
		repaint();
	}
	
	public void updateOpacity() {
		if (isActive() && opacity.directGet() != 1)
			opacity.set(1);
		if (!isActive() && opacity.directGet() != 0)
			opacity.set(0);
	}
	
	public float getThumbLocationWithinBounds(float y, float t) {
		float docHeight = detailView.getHeight();
		if (detailView.getDocument() != null)
			docHeight = detailView.getDocument().getHeight(t);
		return getThumbLocationWithinBounds_(y, docHeight);
	}

	public float getThumbLocationWithinBounds(float y) {
		float docHeight = detailView.getHeight();
		if (detailView.getDocument() != null)
			docHeight = detailView.getDocument().getCurrentHeight();
		return getThumbLocationWithinBounds_(y, docHeight);
	}

	private float getThumbLocationWithinBounds_(float y, float docHeight) {
				
		if (!allowThumbToBeOutside) {
			// try to have the thumb not go further than the current document's height
			if (y + thumbHeight() > doc2bar(docHeight))
				y = doc2bar(docHeight) - thumbHeight();
			// make sure the thumb is inside the bar's bounds
			if (y < barYmin())
				y = barYmin();
			if (y + thumbHeight() > barYmax())
				y = barYmax() - thumbHeight();
		} else {
			final int margin = 4;
			// make sure a small part of the thumb is inside the bar's bounds
			if (y + thumbHeight() < barYmin() + margin)
				y = barYmin() + margin - thumbHeight();
			if (y > barYmax() - margin)
				y = barYmax() - margin;
		}
		return y;
	}
	
	public void moveThumbManually(float y) {
		y = getThumbLocationWithinBounds(y);
		moveThumb(y);
		stabilizer.setViewOfInterest(true);
		detailView.getSpeedNormalizer().setViewOfInterest(true);
		repaint();
		detailView.repaint();
	}
	
	public void translateThumb(float pixels) {
		moveThumbManually(thumbY() + viewH2barH(pixels));
	}
	
	public void pageTranslateThumb(int direction) {
		moveThumbManually(thumbY() + direction * thumbHeight());
	}
	
	public void moveThumbToTop() {
		moveThumbManually(0);
	}
	
	public void moveThumbToBottom() {
		moveThumbManually(barHeight() - thumbHeight());
	}
	
	public void moveThumb(float y) {

		if (isActive()) {
			// update this component
//			thumbY = getThumbLocationWithinBounds(y, detailView.getDocument().getAnimationState());
			thumbY = y;
			// update the detailed view
			detailView.setTranslateY(-bar2doc(thumbY) * detailView.getScale());
		} else {
			// If the scrollbar is not active, do not translate the document.
			thumbY = 0;
			detailView.setTranslateY(0);
		}
		
		updateOpacity();

		repaint();
	}
	
	
	
	/**
	 * Sets the height of the overview panel so it can display all the given text files
	 * FIXME: getTextBlockYPos does not work.
	 * @param files
	 */
	public void fitToFiles(String[] files) {
		
		// used by refitToFileS()
	//	currentFiles = files;
		
		if (files == null || files.length < 2) {
			maxDocumentHeight = detailView.getHeight();
			return;
		}
		
		// 1 -- Compute the maximum document height
		
		float max_y = 0;
		
		// Go through all texts & compute the maximum height
		DocumentAnimator doc = null;
		for (int i=0; i<files.length - 1; i++) {
			// normally, this document is already in cache
			doc = DocumentFactory.createDocument(files[i], files[i+1]);
			doc.setAnimationCoeff(DocumentAnimator.Coeff.ALT);
			doc.setAnimationState(0);
			float y = doc.getCurrentHeight();
			if (y > max_y)
				max_y = y;
		}
		doc.setAnimationState(1);
		float y = doc.getCurrentHeight();
		if (y > max_y)
			max_y = y;

		for (int i=0; i<files.length - 1; i++) {
			doc = DocumentFactory.createDocument(files[i], files[i+1]);
			doc.setAnimationCoeff(DocumentAnimator.Coeff.DEFAULT);
		}

		// 2 -- Update the overview widget

		maxDocumentHeight = max_y;
		
		updateOpacity();
		
		repaint();
	}
	
	public void refitToFiles() {
		String[] currentFiles = detailView.getCurrentFiles();
		if (currentFiles != null)
			fitToFiles(currentFiles);
	}
	
	/**
	 * The usual painting code.
	 */
	public void paint(Graphics g) {
		
ExecTime.start("overview paint");
		
		Graphics2D g2 = (Graphics2D)g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g2.setStroke(THIN_STROKE);
				
		// opacity
		if (opacity.get() <= 0)
			return;
		Composite oldcomposite = null;
		if (opacity.get() < 1) {
			oldcomposite = g2.getComposite();
			g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float)opacity.get()));
		}
		
		// PIX_CAR is not well rendered when antialiasing is activated
		// but xscale is smoother
		g2.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS,
				RenderingHints.VALUE_FRACTIONALMETRICS_ON);
		
		// Paints the background
		g2.setColor(Color.white);
		g2.fillRect(0, 0, getWidth(), getHeight());
		// The border
		g2.setColor(BORDER_COLOR);
		g2.drawRect(PADDING, PADDING, getWidth()-PADDING*2-1, getHeight()-PADDING*2-1);
		
		// Font used for text rendering
//		Font font = new Font("Courier", Font.PLAIN, 14);
//		g2.setFont(font);

		// Shrink to fit
		AffineTransform oldtransform = g2.getTransform();
		g2.translate(PADDING, PADDING);
		float scaley = docH2barH(1);
		float scalex = (getWidth() - 2*PADDING - 1)/(float)detailView.getWidth();
		g2.scale(scalex, scaley);

		// Paint the highlights
		if (detailView.getDocument() != null)
			detailView.getDocument().paint(g2, PaintMode.OVERVIEW, 3/scalex, 3/scaley);
		
		g2.setTransform(oldtransform);
		
		// Paint the thumb
		g2.setColor(THUMB_FILL_COLOR);
		g2.fillRect(0, (int)thumbY() - 1, getWidth(), (int)thumbHeight() + 2);
		g2.setColor(THUMB_COLOR);
		g2.drawRect(0, (int)thumbY() - 1, getWidth() - 1, (int)thumbHeight() + 1);
		g2.drawRect(1, (int)thumbY(), getWidth() - 3, (int)thumbHeight() - 1);

		if (oldcomposite != null)
			g2.setComposite(oldcomposite);
		
ExecTime.end("overview paint");
ExecTime.end("  highlight paint");
ExecTime.end("  text paint");
ExecTime.end("     drawString");

	}

	@Override
	public void mouseClicked(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mouseEntered(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mouseExited(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mousePressed(MouseEvent e) {
		
		stabilizer.setEnabled(false);
		boolean inThumb = e.getY() >= thumbY() - 1 && e.getY() <= thumbY() + thumbHeight() + 1;
		
		if (inThumb || !SwingUtilities.isLeftMouseButton(e)) {
			// Drag
			lastmousex = e.getX();
			lastmousey = e.getY();
			draggingX = false;
			draggingY = true;
			detailView.getTimeline().hoverCursor(true);
		} else {
			// Jump
			if (e.getY() > thumbY())
				moveThumbManually(thumbY() + thumbHeight());
			else
				moveThumbManually(thumbY() - thumbHeight());				
		}
	}

	@Override
	public void mouseDragged(MouseEvent e) {
				
		if (draggingX || draggingY) {
		
			int x = e.getX();

			// Are we now dragging horizontally or vertically?
			if (draggingY) {
				if (x > getWidth() + TIMELINE_DRAG_THRESHOLD || x < - TIMELINE_DRAG_THRESHOLD) {
					detailView.getTimeline().grabCursor();
					draggingY = false;
					draggingX = true;
					constrainedY = e.getLocationOnScreen().y;
					lastmousex = e.getX();
					stabilizer.setEnabled(true);
					stabilizer.setViewOfInterest(true);
					detailView.getSpeedNormalizer().setViewOfInterest(true);
				}
			} else if (draggingX) {
				if (x <= TIMELINE_DRAG_THRESHOLD && x >= 0) {
					detailView.getTimeline().releaseCursor();
					moveThumbManually(thumbY + (e.getY() - lastmousey));
					draggingX = false;
					draggingY = true;
					lastmousey = e.getY();
					stabilizer.setEnabled(false);
				}
			}
			
			if (draggingY) {
			
				float dy = e.getY() - lastmousey;
				if (SwingUtilities.isLeftMouseButton(e)) {
					// translate
					moveThumbManually(thumbY() + dy);
					lastmousey = e.getY();
				} else {
					// zoom
					float newscale = detailView.getScale() * (float)Math.pow(1.01, -dy);
					detailView.setScale(newscale);
					moveThumbManually(thumbY());
					lastmousey = e.getY();
				}
				stabilizer.setViewOfInterest(false);
				detailView.getSpeedNormalizer().setViewOfInterest(false);
				repaint();
				detailView.repaint();

			} else if (draggingX) {
				
				if (x < 0) {
					// to the left
					detailView.getTimeline().dragCursor(e.getX() - lastmousex);
					robot.mouseMove(e.getLocationOnScreen().x, constrainedY);
				} else {
					// to the right
					detailView.getTimeline().dragCursor(e.getX() - lastmousex);
					robot.mouseMove(e.getLocationOnScreen().x, constrainedY);
				}
				lastmousex = e.getX();
			}
		}
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		if (draggingX)
			 detailView.getTimeline().releaseCursor();
		detailView.getTimeline().hoverCursor(false);
		draggingX = false;
		draggingY = false;
		stabilizer.setViewOfInterest(true);
		detailView.getSpeedNormalizer().setViewOfInterest(true);
		stabilizer.setEnabled(true);
		repaint();
		detailView.repaint();
	}

	@Override
	public void mouseMoved(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mouseWheelMoved(MouseWheelEvent e) {
		pageTranslateThumb(e.getWheelRotation());
	}	
	
	public void documentChanged(DocumentAnimator doc) {
		stabilizer.documentChanged(doc);
		updateOpacity();
		repaint();
	}
	
	public float[] getStartEndScrollingVector() {
		return stabilizer.getStartEndScrollingVector();
	}
	
	public float getMaxDocumentHeight() {
		return maxDocumentHeight;
	}
	
	public void setStabilizationEnabled(boolean enable) {
		stabilizer.setEnabled(enable);
	}
}
