/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.diffanimation.render;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Rectangle2D.Float;
import java.util.ArrayList;

import fr.aviz.diffanimation.DiffAnimator;
import fr.aviz.diffanimation.DocumentAnimator;
import fr.aviz.diffanimation.Settings;
import fr.aviz.diffanimation.TextBlock;
import fr.aviz.diffanimation.TextParagraph;
import fr.aviz.diffanimation.Document.Version;
import fr.aviz.diffanimation.DocumentAnimator.PaintMode;
import fr.aviz.diffanimation.TextBlock.TextWord;
import fr.aviz.diffanimation.TextBlock.Type;
import fr.aviz.diffanimation.gui.GUIUtils;
import fr.aviz.utils.ExecTime;

public class TextRenderer {
	
	public static final Color STABLE_HIGHLIGHT_COLOR = null; 
	public static final Color INSERT_HIGHLIGHT_COLOR = new Color(0.65f, 0.95f, 0.65f, 1f); //new Color(0f, 0.8f, 0f, 0.3f);
	public static final Color DELETE_HIGHLIGHT_COLOR = new Color(1f, 0.7f, 0.7f, 1f); //new Color(1f, 0f, 0f, 0.3f);
	public static final Color INSERT_TEXT_COLOR = new Color(0, 0.6f, 0);
	public static final Color DELETE_TEXT_COLOR = new Color(0.7f, 0, 0);
	public static final Color STABLE_HIGHLIGHT_COLOR_OVERVIEW = new Color(0.85f, 0.85f, 0.85f, 1f); 
	public static final Color INSERT_HIGHLIGHT_COLOR_OVERVIEW = new Color(0.27f, 0.85f, 0.27f, 1f);
	public static final Color DELETE_HIGHLIGHT_COLOR_OVERVIEW = new Color(1f, 0.3f, 0.3f, 1f); 
	public static final Color ANCHOR_COLOR = new Color(0, 0, 1, 0.15f);
	public static final float DISPLAY_TEXT_RECTANGLE_BELOW_SCALE = 0.4f;
	
	protected DocumentAnimator docAnimator;
	protected Settings settings;
	
	public TextBlock visibleAnchor;
	float max_y = 0;
	float []geom_t = new float[] {0,0,0,0};
	
	public TextRenderer(DocumentAnimator docAnimator) {
		this.docAnimator = docAnimator;
		this.settings = docAnimator.settings;
	}
		
	/**
	 * Returns the height of the document.
	 * @param motion_t
	 * @return
	 */
	public float updateGeometry(float motion_t) {
		max_y = 0;
		for (TextParagraph tp : docAnimator.getParagraphs()) {
			updateParagraphGeometry(tp, motion_t);
		}
		return max_y;
	}
	
	/**
	 * Compute paragraph geometry.
	 * A paragraph motion is always a linear intepolation between (x0,y0) and (x1,y1)
	 * @param tp
	 * @param motion_t
	 */
	private void updateParagraphGeometry(TextParagraph tp, float motion_t) {
		float []geom_vstart = tp.getPrecomputedGeometry(docAnimator.getStartVersion());
		float motion[] = tp.getPrecomputedMotion(docAnimator.getStartVersion(), docAnimator.getEndVersion());
		
		geom_t[0] = geom_vstart[0]+motion[0]*motion_t;
		geom_t[1] = geom_vstart[1]+motion[1]*motion_t;
		geom_t[2] = geom_vstart[2]+motion[2]*motion_t;
		geom_t[3] = geom_vstart[3]+motion[3]*motion_t;

		float h = settings.lineHeight;
		float w = settings.lineWidth;

		while (geom_t[0] < 0) {
			geom_t[0] += w;
			geom_t[1] -= h;
		}
		while (geom_t[0] >= w) {
			geom_t[0] -= w;
			geom_t[1] += h;
		}
		while (geom_t[2] < 0) {
			geom_t[2] += w;
			geom_t[3] -= h;
		}
		while (geom_t[2] >= w) {
			geom_t[2] -= w;
			geom_t[3] += h;
		}
		tp.setGeometry(geom_t[0],geom_t[1],geom_t[2],geom_t[3]);		
		
		for (TextBlock tb : tp.getBlocks()) {
			updateBlockGeometry(tb, tp, motion_t);
		}
	}
	
	
	
	
   	/**
   	 * Compute block geometry
   	 * @param tb
   	 * @param tp
   	 * @param motion_t
   	 */
	private void updateBlockGeometry(TextBlock tb, TextParagraph tp, float motion_t) {
		tb.computeScales(motion_t, docAnimator.getStartVersion(), docAnimator.getEndVersion());
		
		Version vstart = docAnimator.getStartVersion();
		Version vend = docAnimator.getEndVersion();
		
		float []geom_vstart = tb.getPrecomputedGeometry(vstart);
		float motion[] = tb.getPrecomputedMotion(vstart, vend);
		
		geom_t[0] = geom_vstart[0]+motion[0]*motion_t;
		geom_t[1] = geom_vstart[1]+motion[1]*motion_t;
		geom_t[2] = geom_vstart[2]+motion[2]*motion_t;
		geom_t[3] = geom_vstart[3]+motion[3]*motion_t;
		
		float h = tb.getPrecomputedWrapDownRight(vstart, vend) ? settings.lineHeight : settings.lineHeight*tb.getScaleY();
		float w = settings.lineWidth;
		
		while (geom_t[0] < 0) {
			geom_t[0] += w;
			geom_t[1] -= h;
			geom_t[2] += w;
			geom_t[3] -= h;
		}
		while (geom_t[0] >= w) {
			geom_t[0] -= w;
			geom_t[1] += h;
			geom_t[2] -= w;
			geom_t[3] += h;
		}
		while(geom_t[2] < 0) {
			geom_t[2] += w;
			geom_t[3] -= h;
		}
		/*if (tb.wrapDownRight && geom_t[2] < geom_t[0] && geom_t[3] - geom_t[1] <= 2*h) {
			geom_t[2] += w;
			geom_t[3] -= h;
		}*/
				
		tb.setGeometry(geom_t[0], geom_t[1], geom_t[2], geom_t[3]);
		if (geom_t[3] > max_y)
			max_y = geom_t[3];
		
		for (TextWord tw : tb.getWords()) {
			updateWordGeometry(tw, tb, tp, motion_t);
		}
		
		tb.computeShape(settings, vstart, vend);
	}

	
	private void updateWordGeometry(TextWord tw, TextBlock tb, TextParagraph tp, float motion_t) {
		
		Version vstart = docAnimator.getStartVersion();
		Version vend = docAnimator.getEndVersion();
		
		float []geom_vstart = tw.getPrecomputedGeometry(vstart);
		float motion[] = tw.getPrecomputedMotion(vstart, vend);		
		
		geom_t[0] = geom_vstart[0]+motion[0]*motion_t;
		geom_t[1] = geom_vstart[1]+motion[1]*motion_t;
		geom_t[2] = geom_vstart[2]+motion[2]*motion_t;
		geom_t[3] = geom_vstart[3]+motion[3]*motion_t;
		
		float h = tb.getPrecomputedWrapDownRight(vstart, vend) ? settings.lineHeight : settings.lineHeight*tb.getScaleY();
		float w = settings.lineWidth;
		
		while (geom_t[0] < 0) {
			geom_t[0] += w;
			geom_t[1] -= h;
			geom_t[2] += w;
			geom_t[3] -= h;
		}
		while (geom_t[0] >= w) {
			geom_t[0] -= w;
			geom_t[1] += h;
			geom_t[2] -= w;
			geom_t[3] += h;
		}
		
		tw.setGeometry(geom_t[0],geom_t[1],geom_t[2],geom_t[3]);
	}
	

	
	/**
	 * motion_t is an animation parameter between 0 and 1 used for moving text.
	 * color_t is an animation parameter between 0 and 1 used for animating highlights.
	 * 
	 * @param g
	 * @param motion_t
	 * @param color_t
	 * @param overview
	 */
	public void paint(Graphics2D g, float motion_t, double color_t, PaintMode paintMode, float minHighlightWidth, float minHighlightHeight) {
		if (settings.nominalTextLayout == null)
			return;

		// Extract ymin and ymax from the clip to accelerate rendering
		float ymin = java.lang.Float.MIN_VALUE;
		float ymax = java.lang.Float.MAX_VALUE;
		if (g.getClip() != null) {
			ymin = g.getClip().getBounds().y;
			ymax = ymin + g.getClip().getBounds().height;
		}
		float[] geom;
		
		// Extract scale for text display
		double base_scale = g.getTransform().getScaleX();

ExecTime.start("  highlight paint");
		
		// ------------- DEBUG
		if (DiffAnimator.WORD_BOUNDARIES_DEBUG) {
			for (TextParagraph tp : docAnimator.getParagraphs()) {
				for (TextBlock tb : tp.getBlocks())
					paintWordBoundaries(g, tb, motion_t);
			}
		}
		if (DiffAnimator.BLOCK_BOUNDARIES_DEBUG) {
			for (TextParagraph tp : docAnimator.getParagraphs()) {
				paintBlockBoundaries(g, tp, motion_t);
			}
		}
		if (DiffAnimator.HIGHLIGHTS_BOUNDARIES_DEBUG) {
			for (TextParagraph tp : docAnimator.getParagraphs()) {
				paintHighlightsBoundaries(g, tp, motion_t);
			}
		}
		if (DiffAnimator.PAR_BOUNDARIES_DEBUG) {
			paintParBoundaries(g, motion_t);
		}
		
		// 1 -- Display gray highlights

		if (paintMode == PaintMode.OVERVIEW || paintMode == PaintMode.HISTORYFLOW) {
			for (TextParagraph tp : docAnimator.getParagraphs()) {
				// ignore paragraphs completely outside the view
				geom = tp.getGeometry();
				if (geom[3] < ymin || geom[1] > ymax)
					continue;
				//
				for (TextBlock tb : tp.getBlocks())
					paintHighlights(g, tb, color_t, paintMode, true, 0, 0);
			}
		}

		// 2 -- Display dynamic (red/green) highlights

		if ((paintMode == PaintMode.OVERVIEW && settings.overviewHighlights) || (paintMode == PaintMode.DETAIL && settings.highlights) || paintMode == PaintMode.HISTORYFLOW) {
			for (TextParagraph tp : docAnimator.getParagraphs()) {
				// ignore paragraphs completely outside the view
				geom = tp.getGeometry();
				if (geom[3] < ymin || geom[1] > ymax)
					continue;
				//
				for (TextBlock tb : tp.getBlocks())
					if (tb.getType() != Type.STABLE)
						paintHighlights(g, tb, color_t, paintMode, false, minHighlightWidth, minHighlightHeight);
			}
		}

ExecTime.pause("  highlight paint");
		
		// 3 -- Display text
		
ExecTime.start("  text paint");
		
		if (paintMode == PaintMode.DETAIL) {
			float []twcoord;
			String wordText;
			AffineTransform at;
			Color textColor;
			for (TextParagraph tp : docAnimator.getParagraphs()) {
				// ignore paragraphs completely outside the view
				geom = tp.getGeometry();
				if (geom[3] < ymin || geom[1] > ymax)
					continue;
				//
				for (TextBlock tb : tp.getBlocks())
				{
					textColor = getTextColor(g, tb, color_t);
					boolean displayRectangle = tb.getScalex() * base_scale < DISPLAY_TEXT_RECTANGLE_BELOW_SCALE;
					if (displayRectangle)
						textColor = GUIUtils.multiplyAlpha(textColor, 0.3f);
					g.setColor(textColor);
					
					if (tb == visibleAnchor) {
						paintAnchor(g, tb);
					}

					for(TextWord tw : tb.getWords()) {
						wordText = tw.getText();
						
						twcoord = tw.getGeometry();
						
						// modulo normalization
						while (twcoord[0] < 0) {
							twcoord[0] += settings.lineWidth;
							twcoord[1] -= settings.lineHeight;
							twcoord[2] += settings.lineWidth;
							twcoord[3] -= settings.lineHeight;
						}
						while (twcoord[0] >= settings.lineWidth) {
							twcoord[0] -= settings.lineWidth;
							twcoord[1] += settings.lineHeight;
							twcoord[2] -= settings.lineWidth;
							twcoord[3] += settings.lineHeight;
						}
						
						at = g.getTransform();
						g.translate(twcoord[0], twcoord[1]+settings.baseline*tb.getScaleY());
						g.scale(tb.getScalex(), tb.getScaleY());
						
ExecTime.start("     drawString");
						if (displayRectangle)
							fillRect(g, 0, -settings.ascent*2/3, tw.getVisibleAdvance(), settings.ascent*2/3);
						else
							g.drawString(wordText, 0, 0);
ExecTime.pause("     drawString");
						g.setTransform(at);
													
						if (twcoord[2] > settings.lineWidth) // duplicate the word
							if (twcoord[2] - twcoord[0] < settings.lineWidth) // motion_t != 0 && motion_t != 1)  // do not duplicate a word if doesn't fit in a full line
								{	
									at = g.getTransform();
									g.translate(twcoord[0]-settings.lineWidth, twcoord[1]+settings.baseline*tb.getScaleY()+settings.lineHeight);
									g.scale(tb.getScalex(), tb.getScaleY());
ExecTime.start("     drawString");
									if (displayRectangle)
										fillRect(g, 0, -settings.ascent*2/3, tw.getVisibleAdvance(), settings.ascent*2/3);
									else
										g.drawString(wordText, 0, 0);
ExecTime.pause("     drawString");
									g.setTransform(at);
								}
						}
				}
			}	
		}
		
ExecTime.pause("  text paint");

	}
	
	/**
	 * Sets the current color used to draw
	 * @param g
	 * @param tb the text block to draw
	 * @param  alpha the animation coefficient 
	 */
	protected Color getTextColor(Graphics2D g, TextBlock tb, double alpha) {

		Color color = Color.black;
				 
		if (tb.getType() == Type.INSERT) {
			if (settings.colors) {
				color = INSERT_TEXT_COLOR;
				if (settings.transientHighlights) {
					// fade to black
					float opacity = (float)Math.pow(1-alpha, 0.5f);
					color = GUIUtils.mix(Color.black, color, opacity);
				}
			}
			if (settings.transparency)
				color = GUIUtils.multiplyAlpha(color, (float)alpha);			
		}
		else if (tb.getType() == Type.DELETE) {
			if (settings.colors) {
				color = DELETE_TEXT_COLOR;
				if (settings.transientHighlights) {
					// fade to black
					float opacity = (float)Math.pow(alpha, 0.5f);
					color = GUIUtils.mix(Color.black, color, opacity);
				}
			}
			if (settings.transparency)
				color = GUIUtils.multiplyAlpha(color, 1-(float)alpha);			
 		}
		
		return color;
	}
		
	public void setVisibleAnchor(TextBlock anchor) {
		visibleAnchor = anchor;
	}

	public void paintAnchor(Graphics2D g, TextBlock tb) {
		if (tb == null || tb.getShape().size() == 0)
			return;
		Color oldcol = g.getColor();
		g.setColor(ANCHOR_COLOR);
	    Rectangle2D.Float toprec = tb.getShape().get(0);
	    final float height = 2;
	    fillRect(g, toprec.x, toprec.y, toprec.width, height);
		g.setColor(oldcol);
	}
	
	public void paintHighlights(Graphics2D g, TextBlock tb, double color_t, PaintMode paintMode, boolean gray, float minHighlightWidth, float minHighlightHeight) {
		TextBlock.Type type = tb.getType();

		// Don't display empty stable lines in the overview
		if (tb.getType() == Type.STABLE && tb.getText().trim().length() == 0)
			return;

		boolean fullLine = paintMode == PaintMode.HISTORYFLOW || (paintMode == PaintMode.OVERVIEW && settings.overviewStyle == 1);
		boolean mixGrayAndColor = paintMode == PaintMode.HISTORYFLOW || paintMode == PaintMode.OVERVIEW;
				
		// -- Compute highlight color
		
		Color color = null;
		float opacity = 1;
		final float opacityStrength = 1.3f; // exaggerate opacity if highlights are transient
 		if (type == TextBlock.Type.DELETE) {
			if (mixGrayAndColor && gray)
				color = STABLE_HIGHLIGHT_COLOR_OVERVIEW;
			else {
				color = mixGrayAndColor ? DELETE_HIGHLIGHT_COLOR_OVERVIEW : DELETE_HIGHLIGHT_COLOR;
				if (settings.transientHighlights)
					opacity = (float)(Math.pow(color_t, 0.4f) - Math.pow(color_t, 10f)) * opacityStrength;
			}
 		} else if (type == TextBlock.Type.INSERT) {
			if (mixGrayAndColor && gray)
				color = STABLE_HIGHLIGHT_COLOR_OVERVIEW;
			else {
				color = mixGrayAndColor ? INSERT_HIGHLIGHT_COLOR_OVERVIEW : INSERT_HIGHLIGHT_COLOR;
				if (settings.transientHighlights)
					opacity = (float)(Math.pow(1-color_t, 0.4f) - Math.pow(1-color_t, 10f)) * opacityStrength;
			}
 		} else if (type == TextBlock.Type.STABLE) {
			color = mixGrayAndColor ? STABLE_HIGHLIGHT_COLOR_OVERVIEW : STABLE_HIGHLIGHT_COLOR;
 		}
 		color = GUIUtils.multiplyAlpha(color, opacity);
 		if (opacity == 0)
 			return;
		
 		// -- Paint highlight rectangles
 		
	    ArrayList <Rectangle2D.Float> shape = tb.getShape();
	    for (Rectangle2D.Float r : shape) {
	    	paintHighlightRectangle(g, r.x, r.y, r.width, r.height, color, fullLine, minHighlightWidth, minHighlightHeight);
	    }
	}
	
	/**
	 * Displays a single highlight rectangle.
	 * 
	 * If fullLine is true, the highlight rectangle will take up the whole line
	 * but it will be more transparent (depending on its size relative to the line width)
	 * This is used for painting one-pixel-wide overviews.
	 * 
	 * @param g
	 * @param x
	 * @param y
	 * @param width
	 * @param height
	 * @param color
	 * @param fullLine
	 */
	private void paintHighlightRectangle(Graphics2D g, double x, double y, double width, double height, Color color, boolean fullLine, float minwidth, float minheight) {
		
		Color oldcol = g.getColor();

		if (fullLine) {
			// Make all highlights extend on a whole line. Used to generate the overview.
			final float minOpacity = 0.5f;
			final float opacityStrength = 1.5f;
			float opacity = Math.max(minOpacity, (float)width / (float)settings.lineWidth);
			opacity = Math.min(1, opacity * opacityStrength);
			g.setColor(GUIUtils.multiplyAlpha(color, opacity));
			if (height < minheight) {
				y -= (minheight - height)/2;
				height = minheight;
			}
			fillRect(g, 0, y, settings.lineWidth, height);
			//g.fillRect(0, (int)y, (int)docAnimator.lineWidth, (int)(height+0.5f));
		} else {
			//if (useMinWidth) {
				// Make small blocks more visible
				if (width < minwidth) {
					// x -= (minwidth - width)/2;    // Not necessary
					width = minwidth;
					if (x < 0)
						x = 0;
					if (x + width > settings.lineWidth)
						x = settings.lineWidth - minwidth;
				}
				if (height < minheight) {
					y -= (minheight - height)/2;
					height = minheight;
				}

			//}
			g.setColor(color);
			fillRect(g, x, y, width, height);
			//g.fillRect((int)x, (int)y, (int)(width), (int)(height));
		}
		
		g.setColor(oldcol);

	}
	
	Rectangle2D.Double tmpRect = new Rectangle2D.Double();
	private void fillRect(Graphics2D g, double x, double y, double w, double h) {
		tmpRect.setRect(x, y, w, h);
		g.fill(tmpRect);
	}
	
	/**
	 * For debug
	 */
	public void paintWordBoundaries(Graphics2D g, TextBlock tb, float motion_t) {
		
		Color oldcol = g.getColor();

		for (TextWord tw : tb.getWords()) {
			
		    float[] geom= tw.getGeometry();
			
		    g.setColor(GUIUtils.multiplyAlpha(Color.blue, 0.1f));
		    fillRect(g, geom[0], geom[1], geom[2]-geom[0], geom[3]-geom[1]);
			g.setColor(GUIUtils.multiplyAlpha(Color.blue, 0.6f));
		    g.draw(new Rectangle2D.Float(geom[0], geom[1], geom[2]-geom[0], geom[3]-geom[1]));
		   }
		
	    g.setColor(oldcol);
	}
	
	/**
	 * For debug
	 */
	public void paintBlockBoundaries(Graphics2D g, TextParagraph tp, float motion_t) {
		
		Color oldcol = g.getColor();
		
		for (TextBlock tb : tp.getBlocks()) {
			float[] geom= tb.getGeometry();
			float alpha1 = 0.1f;
			float alpha2 = 0.6f;
			if (tb.getType() == Type.STABLE) {
				alpha1 = 0.05f;
				alpha2 = 0.12f;
			}
			if (geom[2] < geom[0]){
				g.setColor(GUIUtils.multiplyAlpha(Color.magenta, alpha1));
			    g.fill(new Rectangle2D.Float(geom[2], geom[1], geom[0]-geom[2], geom[3]-geom[1]));
				g.setColor(GUIUtils.multiplyAlpha(Color.magenta, alpha2));
			    g.draw(new Rectangle2D.Float(geom[2], geom[1], geom[0]-geom[2], geom[3]-geom[1]));
			   
			}
			else {
		    g.setColor(GUIUtils.multiplyAlpha(Color.pink, alpha1));
		    g.fill(new Rectangle2D.Float(geom[0], geom[1], geom[2]-geom[0], geom[3]-geom[1]));
			g.setColor(GUIUtils.multiplyAlpha(Color.pink, alpha2));
		    g.draw(new Rectangle2D.Float(geom[0], geom[1], geom[2]-geom[0], geom[3]-geom[1]));
		   }
		}
		
	    g.setColor(oldcol);
	}
	
	/**
	 * For debug
	 */
	public void paintParBoundaries(Graphics2D g, float motion_t) {
		
		Color oldcol = g.getColor();

		for (TextParagraph tp : docAnimator.getParagraphs()) {
			
		    float[] geom= tp.getGeometry();
			
		    g.setColor(GUIUtils.multiplyAlpha(Color.green, 0.1f));
		    g.fill(new Rectangle2D.Float(geom[0], geom[1], geom[2]-geom[0], geom[3]-geom[1]));
			g.setColor(GUIUtils.multiplyAlpha(Color.green, 0.6f));
		    g.draw(new Rectangle2D.Float(geom[0], geom[1], geom[2]-geom[0], geom[3]-geom[1]));
		   }
		
	    g.setColor(oldcol);
	}
	
	public void paintHighlightsBoundaries(Graphics2D g, TextParagraph tp, float motion_t) {		
		Color oldcol = g.getColor();

		for (TextBlock tb : tp.getBlocks()) {
			ArrayList<Float>shape = tb.getShape();
			for (Rectangle2D.Float r : shape) {
				g.setColor(GUIUtils.multiplyAlpha(Color.orange, 0.1f));
				g.fill(r);
				g.setColor(GUIUtils.multiplyAlpha(Color.orange, 0.6f));
				g.draw(r);
			}	
		}
		g.setColor(oldcol);
	}
}
