/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.diffanimation.timeline;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseEvent;
import java.util.ArrayList;

import javax.swing.JComponent;
import javax.swing.SwingUtilities;
import javax.swing.event.MouseInputListener;

import fr.aviz.animation.AnimationListener;
import fr.aviz.animation.AnimationTimer;
import fr.aviz.diffanimation.DiffAnimator;
import fr.aviz.diffanimation.DocumentAnimator;
import fr.aviz.diffanimation.gui.GUIUtils;
import fr.aviz.diffanimation.overview.DiffOverview;
import fr.aviz.utils.DocumentFactory;

public class DiffTimeline extends JComponent implements MouseInputListener {

	// The component's vertical and horizontal internal margins
	public static final int PADDING = 2;

	public static final int MIN_MARK_WIDTH = 1;
	public static final int MAX_MARK_WIDTH = 3;	
	public static final int MARK_MARGIN = 2;

	// Colors
	public static final Color MARK_COLOR = new Color(0.7f, 0.7f, 0.7f);
	public static final Color BORDER_COLOR = DiffOverview.BORDER_COLOR;
	public static final Color CURSOR_COLOR = DiffOverview.THUMB_COLOR;
	public static final Color CURSOR_FILL_COLOR = DiffOverview.THUMB_FILL_COLOR;
	public static final Color HOVER_COLOR = GUIUtils.multiplyAlpha(DiffOverview.THUMB_COLOR, 0.4f);
	public static final Color HOVER_FILL_COLOR = new Color(0.9f, 0.9f, 1, 0.8f);;
	public static Stroke THIN_STROKE = new BasicStroke(1);
	
	static final boolean regularDistribution = false; 

	final DiffAnimator detailView;
	ArrayList<RevisionMark> marks = new ArrayList<RevisionMark>();
	ArrayList<HistoryFlow> flows = new ArrayList<HistoryFlow>();
	ArrayList<Float> distances = new ArrayList<Float>();
	float pickWidth;
	float maxDocumentHeight;
	
	RevisionMark hoveredMark;
	RevisionMark focusedMark;
	RevisionMark cursor = new RevisionMark(null);
	RevisionMark mark0, mark1;
	HistoryFlow flow01 = null;
	int mouse_x;
	boolean dragging_cursor = false;
	boolean useShortestEditionPath = false;
	boolean active = true;
	
	DiffTimeline layoutModel = null;
	
	Thread shortestPathComputationThread = null;
	
	public DiffTimeline(final DiffAnimator detailView, int height) {
		super();
		this.detailView = detailView;
		cursor.setX(-1000);
		hoveredMark = null;
		setPreferredSize(new Dimension(800, height));
		addMouseListener(this);
		addMouseMotionListener(this);
		addComponentListener(new ComponentAdapter() {
			public void componentResized(ComponentEvent e) {
				redrawHistoryFlows();
				layoutMarks();
			}
		});
		/*AnimationTimer.addListener(new AnimationListener() {
			public void animateEvent() {
				updateCursorPosition();
			}
		});*/
	}
	
	public void useShortestEditionPath(boolean enable) {
		this.useShortestEditionPath = enable;
	}
	
	public void setLayoutModel(DiffTimeline timeline) {
		this.layoutModel = timeline;
	}
	
	public void clear() {
		mark0 = null;
		mark1 = null;
		flow01 = null;
		hoveredMark = null;
		focusedMark = null;
		cursor.setX(-1000);
		marks.clear();
		flows.clear();
		distances.clear();
	}
	
	public void newFiles() {
		
		clear();
		
		String[] files = useShortestEditionPath ? detailView.getCurrentFilesShortestPath() : detailView.getCurrentFiles();
		
		if (files == null && useShortestEditionPath) {
			shortestPathComputationThread = new Thread() {
				public void run() {
					detailView.computeShortestPath();
					if (shortestPathComputationThread == this) {
						newFiles();
						if (layoutModel != null && layoutModel.active)
							hideRedundantMarks(layoutModel);
					}
				}
			};
			shortestPathComputationThread.start();
			return;
		}
		
		if (files == null || files.length < 2)
			return;
		
		for (int i=0; i<files.length; i++) {
			RevisionMark mark = new RevisionMark(files[i]);
			marks.add(mark);
			if (i > 0) {
				flows.add(new HistoryFlow(marks.get(i-1), mark));
				float changeMagnitude = DocumentFactory.createDocument(files[i-1], files[i]).getChangeMagnitude();
				distances.add(changeMagnitude);
			}
		}
				
		layoutMarks();
		repaint();
	}
	
	public void setMaxDocumentHeight(float maxDocumentHeight) {
		this.maxDocumentHeight = maxDocumentHeight;
//		redrawHistoryFlows();
	}
	
	public void layoutMarks() {
		
		if (marks.size() <= 1)
			return;
		
		int w = getWidth();
		int h = getHeight();
		int n = marks.size();
		int markWidth = w / n / 8;
		markWidth = Math.min(markWidth, MAX_MARK_WIDTH);
		markWidth = Math.max(markWidth, MIN_MARK_WIDTH);
		pickWidth = markWidth * 8;
		int internal_w = w - markWidth - 4; // includes selection width
		float x;
		
		if (layoutModel != null) {
			for (int i=0; i<n; i++) {
				RevisionMark model = layoutModel.marks.get(layoutModel.getMarkIndex(marks.get(i).filename));
				marks.get(i).setBounds(model.bounds.x, model.bounds.y, model.bounds.width, model.bounds.height);
			}
		} else if (regularDistribution) {
			for (int i=0; i<n; i++) {
				x = 2 + i * internal_w / (n-1);
				marks.get(i).setBounds(x, 2, markWidth, h-4);
			}
		} else {
			float totaldistance = 0;
			for (int i=0; i<n-1; i++) {
				totaldistance += 0.5f + distances.get(i);
			}
			x = 2;
			for (int i=0; i<n; i++) {
				if (i != 0) {
					x += (0.5f + distances.get(i-1)) * internal_w / totaldistance;
				}
				marks.get(i).setBounds(x, 2, markWidth, h-4);
			}
		}
		
		for (HistoryFlow flow : flows)
			flow.layout();
		if (flow01 != null)
			flow01.layout();
		cursor.setBounds(-1000, 2, markWidth, h-4);
		updateCursorPosition();
		redrawHistoryFlows();
	}
	
	public void setActive(boolean active) {
		this.active = active;
		if (!active) {
			cursor.setX(-1000);
			hoveredMark = null;
			focusedMark = null;
			mark0 = null;
			mark1 = null;
			flow01 = null;
		}
		repaint();
	}

	public void setAllVisible(boolean visible) {
		for (RevisionMark m : marks)
			m.setVisible(visible);
		for (HistoryFlow f : flows)
			f.setVisible(visible);
		repaint();
	}
	
	public void hideRedundantMarks(DiffTimeline timeline) {
		HistoryFlow previousFlow = null;
		for (HistoryFlow f : flows) {
			if (timeline.getFlow(f.mark0.filename, f.mark1.filename) != null) {
				f.setVisible(false);
				if (previousFlow == null || !previousFlow.visible)
					marks.get(getMarkIndex(f.mark0.filename)).setVisible(false);
			}
			previousFlow = f;
		}
		if (previousFlow != null && !previousFlow.visible)
			marks.get(marks.size()-1).setVisible(false);
		repaint();
	}
	
	protected HistoryFlow getFlow(String file1, String file2) {
		for (HistoryFlow flow : flows) {
			if (flow.mark0.filename.equals(file1) && flow.mark1.filename.equals(file2)) {
				return flow;
			}
		}
		return null;
	}
	
	protected void redrawHistoryFlows() {
		for (HistoryFlow flow : flows)
			flow.newBitmap(maxDocumentHeight);
		if (flow01 != null)
			flow01.newBitmap(maxDocumentHeight);
		drawing_flow = 0;
		drawFlowsIteration();	
	}
	
	int drawing_flow = 0;
	
	protected void drawFlowsIteration() {
		if (drawing_flow > flows.size())
			return;
		
		HistoryFlow currentFlow;
		if (drawing_flow == flows.size()) {
			if (flow01 != null) {
				currentFlow = flow01;
			} else
				return;
		} else {
			currentFlow = flows.get(drawing_flow);
		}
		
		// Draw
		/*if (layoutModel != null) {
			HistoryFlow sameFlow = layoutModel.getFlow(currentFlow.mark0.filename, currentFlow.mark1.filename);
			currentFlow.setBitmap()
		}
		*/
		if (currentFlow.drawBitmapIteration())
			drawing_flow++;
		
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				drawFlowsIteration();
			}
		});
		repaint();
	}
	
	protected int getMarkIndex(String filename) {
		for (int i=0; i<marks.size(); i++) {
			if (marks.get(i).filename.equals(filename))
				return i;
		}
		return -1;
	}
	
	protected int getMarkIndex(RevisionMark mark) {
		for (int i=0; i<marks.size(); i++) {
			if (marks.get(i) == mark)
				return i;
		}
		return -1;
	}

	protected RevisionMark pick(int x) {
		RevisionMark bestPick = null;
		float minDistance = Float.MAX_VALUE;
		if (cursor.contains(x, pickWidth)) {
			minDistance = Math.abs(cursor.getX() - x);
			bestPick = cursor; 
		}
		for (RevisionMark mark : marks) {
			if (mark.contains(x, pickWidth)) {
				if (Math.abs(mark.getX() - x) < minDistance) {
					minDistance = Math.abs(mark.getX() - x);
					bestPick = mark;
				}
			}
		}
		return bestPick;
	}
	
	public void updateCursorPosition() {
		
		if (!active)
			return;
		
		DocumentAnimator document = detailView.getDocument();
		
		if (document == null) {
			cursor.translate(-1000);
			repaint();
			return;
		}
		
		int start = getMarkIndex(document.getFilename1());
		int end = getMarkIndex(document.getFilename2());
		if (start == -1)
			start = end;
		if (end == -1)
			end = start;
		if (start != -1 && end != -1) {
			RevisionMark newmark0 = marks.get(start);
			RevisionMark newmark1 = marks.get(end);
			if (newmark0 != mark0 || newmark1 != mark1) {
				mark0 = newmark0;
				mark1 = newmark1;
				// Draw the new history flow
				if (getMarkIndex(mark1) - getMarkIndex(mark0) > 1) {
					flow01 = new HistoryFlow(mark0, mark1);
					flow01.layout();
					flow01.newBitmap(maxDocumentHeight);
					drawing_flow = flows.size();
					drawFlowsIteration();	
				} else {
					flow01 = null;
				}
			}
			float t = document.getAnimationState();
			float oldCursorPosition = cursor.getX();
			float newCursorPosition = getX0() + (getX1() - getX0()) * t;
			if (oldCursorPosition != newCursorPosition) {
				repaintCursor();
				cursor.setX(newCursorPosition);
				if (!dragging_cursor)
					updateHover();
				repaintCursor();
			}
		} else {
			mark0 = null;
			mark1 = null;
			flow01 = null;
			cursor.setX(-1000);
			repaint();
		}
	}

	protected void repaintCursor() {
		repaint((int)cursor.border.x-2-8, (int)cursor.border.y-2, (int)cursor.border.width+5+8+8, (int)cursor.border.height+5);
	}
	
	protected void updateHover() {
		updateHover(mouse_x);
	}
	
	protected float getX0() {
		if (mark0 != null)
			return mark0.getX();
		return -2;
	}
	
	protected float getX1() {
		if (mark1 != null)
			return mark1.getX();
		return -1;
	}

	protected void updateHover(int x) {
		if (focusedMark != null)
			return;
		RevisionMark mark = pick(x);
		if (mark != hoveredMark) {
			hoveredMark = mark;
			repaint();
			if (hoveredMark == null)
				detailView.updateWindowTitle();
			else
				detailView.setWindowTitle(hoveredMark.getFilename());
		}
		mouse_x = x;
	}

	public void paint(Graphics g_) {
		
		if (marks.size() == 0)
			return;
		
		Graphics2D g = (Graphics2D)g_;
		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
		g.setStroke(THIN_STROKE);

		// Background
		g.setColor(Color.white);
		g.fillRect(0, 0, getWidth(), getHeight());
		// Border
//		g.setColor(BORDER_COLOR);
//		g.drawRect(PADDING, PADDING, getWidth()-PADDING*2-1, getHeight()-PADDING*2-1);
		
		String[] files = detailView.getCurrentFiles();
		if (files == null || files.length < 2)
			return;
		
		// History flow
		for (HistoryFlow flow : flows) {
//			boolean skipped = (mark.getX() > getX0() && mark.getX() < getX1());
			flow.paint(g);
		}
		if (flow01 != null)
			flow01.paint(g);
		
		// Revision marks
		for (RevisionMark mark : marks) {
			boolean skipped = (mark.getX() > getX0() && mark.getX() < getX1());
			mark.paint(g, skipped);
		}
		
		// Hovered & focus marks
		if (hoveredMark != null && hoveredMark != cursor) {
			//boolean skipped = (hoveredMark.getX() > getX0() && hoveredMark.getX() < getX1());
			hoveredMark.paintBorder(g, false, HOVER_FILL_COLOR, HOVER_COLOR);
		}
		if (focusedMark != null) {
			//boolean skipped = (focusedMark.getX() > getX0() && focusedMark.getX() < getX1());
			focusedMark.paintBorder(g, false, HOVER_FILL_COLOR, HOVER_COLOR);
		}
		
		// Cursor
		if (cursor.getX() >= 0) {
			if (cursor != hoveredMark)
				cursor.paintBorder(g, false, CURSOR_FILL_COLOR, CURSOR_COLOR);
			else {
				if (!dragging_cursor) {
					cursor.paintBorder(g, false, CURSOR_FILL_COLOR, CURSOR_COLOR);
					float t = (cursor.getX() - getX0()) / (getX1() - getX0());
					if (t > 0.1f)
						cursor.paintLeftArrow(g, HOVER_FILL_COLOR, HOVER_COLOR);
					if (t < 0.9f)
						cursor.paintRightArrow(g, HOVER_FILL_COLOR, HOVER_COLOR);
				} else {
					cursor.paintBorder(g, false, CURSOR_FILL_COLOR, CURSOR_COLOR);
					//cursor.paintArrows(g, CURSOR_FILL_COLOR, CURSOR_COLOR);
				}
			}
		}
	}
	
	@Override
	public void mouseEntered(MouseEvent e) {
		if (marks.size() == 0)
			return;
		
		updateHover(e.getX());
	}
	
	@Override
	public void mouseMoved(MouseEvent e) {
		if (marks.size() == 0)
			return;
		
		focusedMark = null;
		updateHover(e.getX());
	}
	
	@Override
	public void mouseExited(MouseEvent e) {
		if (marks.size() == 0)
			return;
		
		updateHover(-1000);
	}

	@Override
	public void mousePressed(MouseEvent e) {
		
		if (marks.size() == 0)
			return;
		
		if (getParent() instanceof CompoundDiffTimeline && !active)
			((CompoundDiffTimeline)getParent()).setActiveTimeline(this);
		
		if (hoveredMark != null) {
			if (hoveredMark != cursor) {
				if (hoveredMark.filename.equals(detailView.getLastFilename())) {
					cursor.setX(hoveredMark.getX());
				} else {
					detailView.queueAnimation(hoveredMark.getFilename());
				}
			} else {
				dragging_cursor = true;
				detailView.emptyAnimationQueue();
				detailView.boostCurrentAnimation(false);
			}
			repaint();
			mouse_x = e.getX();
		}
	}
	
	@Override
	public void mouseDragged(MouseEvent e) {
		
		if (marks.size() == 0)
			return;
	
		if (dragging_cursor) {
			int dx = e.getX() - mouse_x;
			mouse_x = e.getX();
			cursor.translate(dx);
			float t = (mouse_x - getX0()) / (getX1() - getX0());
			detailView.setAnimationState(t, true);
			repaint();
		} else {
			// crossing-based interaction
			RevisionMark lastHover = hoveredMark;
			updateHover(e.getX());
			if (hoveredMark != null && hoveredMark != lastHover && hoveredMark.getFilename() != null) {
				detailView.queueAnimation(hoveredMark.getFilename());
			}
		}
	}
	
	@Override
	public void mouseClicked(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		
		if (marks.size() == 0)
			return;

		// TODO Auto-generated method stub
		if (dragging_cursor) {
			float t = (mouse_x - getX0()) / (getX1() - getX0());
			String file1 = detailView.getDocument().getFilename1();
			String file2 = detailView.getDocument().getFilename2();
			if (t > 0.5) {
				detailView.startAnimation(file1, file2, t, 1);
			} else {
				detailView.startAnimation(file1, file2, t, 0);
			}
			detailView.boostCurrentAnimation(true);
			dragging_cursor = false;
			repaint();
		}
	}
	
	public void advance(int direction) {
		
		if (!dragging_cursor) {
			int lastMark = getMarkIndex(detailView.getLastFilename());
			if (lastMark != -1) {
				int newMark = lastMark + direction;
				if (newMark >= 0 && newMark < marks.size()) {
					focusedMark = marks.get(newMark);
					hoveredMark = null;
					detailView.queueAnimation(focusedMark.getFilename());
					repaint();
				}
			}
		}
	}
	
    public void hoverCursor(boolean hover) {
		// Quick and dirty to control the cursor: we emulate mouse events.
    	if (hover)
    		mouseMoved(new MouseEvent(this, 0, 0, 0, Math.round(cursor.getX()), getHeight()/2, 0, false));
    	else
    		mouseMoved(new MouseEvent(this, 0, 0, 0, -1000, getHeight()/2, 0, false));
	}
	
	public void grabCursor() {
		// Quick and dirty to control the cursor: we emulate mouse events.		
		mousePressed(new MouseEvent(this, 0, 0, 0, Math.round(cursor.getX()), getHeight()/2, 0, false));
	}
	
	public void dragCursor(int dx) {
		// Quick and dirty to control the cursor: we emulate mouse events.		
		mouseDragged(new MouseEvent(this, 0, 0, 0, Math.round(cursor.getX()+dx), getHeight()/2, 0, false));
	}
	
	public void releaseCursor() {
		// Quick and dirty to control the cursor: we emulate mouse events.		
		mouseReleased(new MouseEvent(this, 0, 0, 0, Math.round(cursor.getX()), getHeight()/2, 0, false));
	}
}
