/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.utils;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.text.DecimalFormat;
import java.util.Random;

import javax.swing.JOptionPane;
import javax.swing.SwingWorker;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.clapper.util.html.HTMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

import fr.aviz.diff.DiffChecker;
import fr.aviz.diff.methods.PonsDiff;
import fr.aviz.diffanimation.gui.JLoadDirectory;

public class QueryAPITask extends SwingWorker<Void, Void> {
	private String outputDir;
	private String wikiversion;
	private String article;
	private String url;
	private JLoadDirectory loader;
	private int progress;
	
	/**
	 * If true, will include links in the plain text (uses a web service).
	 */
	static final boolean INCLUDE_LINKS_IN_TEXT = false; 
	
	static final DecimalFormat REVISION_FORMAT = new DecimalFormat("00");
	
	//
	private static DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
	private static DocumentBuilder docBuilder;
	static {
		try {docBuilder = docBuilderFactory.newDocumentBuilder();} catch (Exception e) {e.printStackTrace();}
	}
	
	public void setSettings(String dir, String wikiversion, String article, String url, JLoadDirectory loader) {
		this.outputDir = dir;
		this.wikiversion = wikiversion;
		this.article = article;
		this.url = url;
		this.loader = loader;
	}
	
	/*
	 * Main task. Executed in background thread.
	 */
	@Override
	public Void doInBackground() {
		progress = 0;
		//Initialize progress property.
		setProgress(0);
		
		// Collect revisions IDs from API
		int[] revisions = getRevisionIDs(url);
		System.out.println("Got revisions: " + revisions.length);
		
		// Download articles
		int nbrevs = downloadHistoryFromUrl(revisions);
		
		// Load in Diffamation
		if (nbrevs > 2) {
			File file = new File(outputDir+"/" + article);
		 	loader.loadFiles(file);
		} 
		 else 
			 JOptionPane.showMessageDialog(null, "Not enough revisions to show", "Wikipedia API result", JOptionPane.WARNING_MESSAGE);
		
		setProgress(100);
		return null;
	}

	/*
	 * Executed in event dispatching thread
	 */
	@Override
	public void done() {
		//startButton.setEnabled(true);
		//setCursor(null); //turn off the wait cursor
		//taskOutput.append("Done!\n");
	}
	
	
	protected int[] getRevisionIDs(String url) {
		try {
			setProgress(5);
			progress = 5;
			Document doc = docBuilder.parse(url);
			NodeList revList = doc.getElementsByTagName("rev");
			int[] revs = new int[revList.getLength()];
			for (int i=0; i<revList.getLength(); i++) {
				String rev = revList.item(i).getAttributes().getNamedItem("revid").getNodeValue();
				revs[i] = Integer.parseInt(rev);
			}
			setProgress(10);
			progress = 10;
			return revs;
		} catch (Exception e) {
			JOptionPane.showMessageDialog(null, "Wikipedia API Error\n" + url, "Wikipedia API query", JOptionPane.ERROR_MESSAGE);
			//e.printStackTrace();
			return new int[]{};
		}
	}
	
	
	private int downloadHistoryFromUrl(int[]revs) {
		String previous_s = null;
		int increment = ((100-10) / revs.length) / 3;
		
		for (int i=0; i<revs.length; i++) {

			String filename = outputDir + "/" + article + REVISION_FORMAT.format(revs.length - i) + ".txt";

			// Get the article
			System.out.print("Retrieving " + filename + " ");
			String s = getPlainTextArticleFromID(wikiversion, revs[i]);
			progress+=increment;
			setProgress(progress);
			
			if (previous_s != null) {
				PonsDiff diff = new PonsDiff(s, previous_s);
				if (diff.getInsertCount() + diff.getDeleteCount() == 0)
					continue;
			}
			progress+=increment;
			setProgress(progress);
			
			// Save it into a file
			try {
				BufferedWriter out = new BufferedWriter(new FileWriter(filename));
				//				someText.replaceAll("\n", System.getProperty("line.separator"));
				out.write(s);
				out.close();
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			previous_s = s;
			progress+=increment;
			setProgress(progress);
		}
		
		return revs.length;
	}
	
	protected static String getPlainTextArticleFromID(String wikiversion, int articleId) {
		String url = wikiversion + "/w/index.php?oldid=" + articleId + "&action=render";
		return getPlainTextArticleFromURL(url);
	}
	
	/**
	 * 
	 */
	private static String getPlainTextArticleFromURL(String url) {
		try {
			if (INCLUDE_LINKS_IN_TEXT)
				url = "http://cgi.w3.org/cgi-bin/html2txt?url=" + URLEncoder.encode(url, "UTF-8") + "&noinlinerefs=on&endrefs=on";		
			String text;
			// -- Clean up
			if (INCLUDE_LINKS_IN_TEXT) {
				// The html is already plain text now
				text = getContent(url);
			} else {
				
				String html = getContent(url);
				
				html = removeElement(html, "table", null, "infobox");
				html = removeElement(html, "table", null, "navbox");
				html = removeElement(html, "table", "toc", null);

				// Remove [edit] sections in the html (also removes the extra space)
				html = html.replaceAll("<span class=\\\"editsection\\\">.*?<\\/span> ", "");
				
				// Convert
				text = HTMLUtil.textFromHTML(html);
				// Clean up line breaks in the text
				for (int i=0; i<10; i++)
					text = text.replace("\n  \n", "\n");
				for (int i=0; i<10; i++)
					text = text.replace("\n \n", "\n");
				for (int i=0; i<10; i++)
					text = text.replace("\n\n", "\n");
				text = text.replace("\n", "\n\n");
			}
			return text;
		} catch (Exception e) {
			e.printStackTrace();
			return e.getMessage();
		}
	}
	
	private static String getContent(String url) {
		try {
			System.out.print("\t.");
			URLConnection yc = (new URL(url)).openConnection();
			BufferedReader in = new BufferedReader(new InputStreamReader(yc.getInputStream()));
			String html = "";
			String inputLine;
			int nlines = 0;
			while ((inputLine = in.readLine()) != null) {
				nlines ++;
				if (nlines%50 == 0)
					System.out.print(".");
				html += inputLine + "\n";
			}
			in.close();
			System.out.println();
			return html;
		} catch (Exception e) {
			e.printStackTrace();
			return "";
		}
	}
	
	
	/**
	 * This ugly method is there to clean up HTML without having to parse the XML. XML parsers
	 * need well-formed HTML and wikipedia does not provide this. 
	 */
    public static String removeElement(String html, String element, String id, String cclass) {
    	String open = "<" + element;
    	if (id != null)
    		open += " id=\"" + id + "\"";
    	if (cclass != null)
    		open += " class=\"" + cclass;// + "\"";
    	String close = "</" + element + ">";
   	
        while (html.indexOf(open) != -1) {
        	int o_start = html.indexOf(open);
        	int o_end = 0;
        	
        	int tab = 1;
        	int from = o_start + open.length();
        	while (tab != 0) {
	        	int o = bestSearch(html.indexOf("<" + element + " ", from), html.indexOf("<" + element + ">", from));
	        	int c = html.indexOf(close, from);
	        	int best = bestSearch(o, c);
	        	if (c == -1)
	        		return null;
	        	if (best == o) {
	        		tab++;
	        		from = o + element.length() + 2;
	        	} else if (best == c) {
	        		tab--;
	        		from = c + element.length() + 2;
	        	} else
	        		return null;
	        	if (tab == 0) {
	        		o_end = c + close.length();
	        	}
        	}
//System.err.println("Removed " + element);
        			
        	html = html.substring(0, o_start) + html.substring(o_end);
        }
        
        return html;
    }
    
    private static int bestSearch(int i1, int i2) {
    	if (i1 == -1)
    		return i2;
    	if (i2 == -1)
    		return i1;
    	return Math.min(i1, i2);
    }
}