/* 
 * Copyright (c) 2010 Microsoft Corporation and Inria. 
 * Contributors: Fanny Chevalier and Pierre Dragicevic. 
 * 
 * You may distribute under the terms of either the CeCILL-B License or the 
 * CeCILL version 2 License.
 */
package fr.aviz.utils;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.text.DecimalFormat;
import java.util.ArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.clapper.util.html.HTMLUtil;
import org.w3c.dom.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import fr.aviz.diff.DiffChecker;
import fr.aviz.diff.methods.PonsDiff;

public class WikipediaDownloader {

	/**
	 * The title of the wikipedia article to download
	 */
	static final String ARTICLE_TITLE = "Portal:Pope/Intro";

	/**
	 * If true, will include links in the plain text (uses a web service).
	 */
	static final boolean INCLUDE_LINKS_IN_TEXT = false; 
	/**
	 * Max number of revisions, starting from the most recent one.
	 */
	public static final String OUTPUT_DIR_PREFIX = "./data/wikipedia/";
	static final DecimalFormat REVISION_FORMAT = new DecimalFormat("00");
	
	//
	private static DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
	private static DocumentBuilder docBuilder;
	static {
		try {docBuilder = docBuilderFactory.newDocumentBuilder();} catch (Exception e) {e.printStackTrace();}
	}

	public static void main(String[] args) {
//		downloadRandomArticles(1000, 1500, 5000, 5, 50);
//		downloadRandomArticles(50, 1500, 2, 2);
	}

	public static void downloadRandomArticles(String outputDir, boolean clearFolder, int articleCount, int minLength, int maxLength, int minRevisions, int maxRevisions) {
		// clear folder
		if (clearFolder)
			DiffChecker.clearFolder(outputDir);

		int articlesTried = 0;
		
		for (int i=0; i<articleCount; i++) {
			boolean ok;
			do {
				ok = false;
				String title = getRandomArticleTitle();
				String article = getPlainTextArticleFromTitle(title);
				// Article length must be >= 'minLength'
				if (article.length() >= minLength) {
					int[] IDs = getRevisionIDs(title, maxRevisions);
					// Number of revisions must be >= 'revisions'
					if (IDs.length >= Math.min(minRevisions, maxRevisions)) {
						downloadHistory(outputDir, title, maxRevisions, true);
						ok = true;
					}
				}
				articlesTried++;
			} while (!ok);
			System.out.println("Kept " + ((i+1) * 100 / articlesTried) + " % of all articles so far");
		}
	}

	
	
	public static void downloadHistory(String outputDir, String article, String url, boolean pruneIdenticalRevisions) {
		// Create output directory
		(new File(outputDir)).mkdirs();

		// clean the string for the output filename
		String articlef = article.replace("/", "_");
		articlef = articlef.replace(":", "_");
		articlef = articlef.replace(".", "_");

		int[] revs = getRevisionIDs(url);

		System.out.println("Got " + revs.length + " revisions for " + article);

		long time0 = System.currentTimeMillis();
		String previous_s = null;

		// Show the progress bar
		
		for (int i=0; i<revs.length; i++) {

			String filename = outputDir + "/" + articlef + REVISION_FORMAT.format(revs.length - i) + ".txt";

			// Get the article
			System.out.print("Retrieving " + filename + " ");
			if (i > 0) {
				int estimatedTimeLeft = (int)((System.currentTimeMillis() - time0) * (revs.length - 1) / (float)i);
				System.out.print(" (" + estimatedTimeLeft / 1000 / 60 + " minutes left)  ");
			}
			String s = getPlainTextArticleFromID(revs[i]);

			if (pruneIdenticalRevisions && previous_s != null) {
				PonsDiff diff = new PonsDiff(s, previous_s);
				if (diff.getInsertCount() + diff.getDeleteCount() == 0)
					continue;
			}
			
			// Save it into a file
			try {
				BufferedWriter out = new BufferedWriter(new FileWriter(filename));
				//				someText.replaceAll("\n", System.getProperty("line.separator"));
				out.write(s);
				out.close();
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			previous_s = s;
		}
	}
	
	
	public static void downloadHistory(String outputDir, String article, int maxRevisions, boolean pruneIdenticalRevisions) {

		// Create output directory
		(new File(outputDir)).mkdirs();

		// clean the string for the output filename
		String articlef = article.replace("/", "_");
		articlef = articlef.replace(":", "_");
		articlef = articlef.replace(".", "_");

		int[] revs = getRevisionIDs(article, maxRevisions);

		System.out.println("Got " + revs.length + " revisions for " + article);

		long time0 = System.currentTimeMillis();
		String previous_s = null;

		// Show the progress bar
		
		for (int i=0; i<revs.length; i++) {

			String filename = outputDir + "/" + articlef + REVISION_FORMAT.format(revs.length - i) + ".txt";

			// Get the article
			System.out.print("Retrieving " + filename + " ");
			if (i > 0) {
				int estimatedTimeLeft = (int)((System.currentTimeMillis() - time0) * (revs.length - 1) / (float)i);
				System.out.print(" (" + estimatedTimeLeft / 1000 / 60 + " minutes left)  ");
			}
			String s = getPlainTextArticleFromID(revs[i]);

			if (pruneIdenticalRevisions && previous_s != null) {
				PonsDiff diff = new PonsDiff(s, previous_s);
				if (diff.getInsertCount() + diff.getDeleteCount() == 0)
					continue;
			}
			
			// Save it into a file
			try {
				BufferedWriter out = new BufferedWriter(new FileWriter(filename));
				//				someText.replaceAll("\n", System.getProperty("line.separator"));
				out.write(s);
				out.close();
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			previous_s = s;
		}

	}

	
	/**
	 * Gets all available revisions from the query
	 * @param url the query
	 * @return
	 */
	protected static int[] getRevisionIDs(String url) {
		try {
			Document doc = docBuilder.parse(url);
			NodeList revList = doc.getElementsByTagName("rev");
			int[] revs = new int[revList.getLength()];
			for (int i=0; i<revList.getLength(); i++) {
				String rev = revList.item(i).getAttributes().getNamedItem("revid").getNodeValue();
				revs[i] = Integer.parseInt(rev);
			}
			return revs;
		} catch (Exception e) {
			e.printStackTrace();
			return new int[]{};
		}
	}
	
	/**
	 * 
	 */
	protected static int[] getRevisionIDs(String articleTitle, int max) {
		try {
			String url = "http://en.wikipedia.org/w/api.php?action=query&prop=revisions&titles=" + articleTitle + "&rvprop=ids&format=xml&rvlimit=" + max;
			Document doc = docBuilder.parse(url);
			NodeList revList = doc.getElementsByTagName("rev");
			int[] revs = new int[revList.getLength()];
			for (int i=0; i<revList.getLength(); i++) {
				String rev = revList.item(i).getAttributes().getNamedItem("revid").getNodeValue();
				revs[i] = Integer.parseInt(rev);
			}
			return revs;
		} catch (Exception e) {
			e.printStackTrace();
			return new int[]{};
		}
	}
	
	
	protected static int[] getRevisionIDsFrom(String articleTitle, int max, int from) {
		try {
			String url = "http://en.wikipedia.org/w/api.php?action=query&prop=revisions&titles=" + articleTitle + "&rvprop=ids&format=xml&rvlimit=" + max;
			Document doc = docBuilder.parse(url);
			NodeList revList = doc.getElementsByTagName("rev");
			int[] revs = new int[revList.getLength()];
			for (int i=0; i<revList.getLength(); i++) {
				String rev = revList.item(i).getAttributes().getNamedItem("revid").getNodeValue();
				revs[i] = Integer.parseInt(rev);
			}
			return revs;
		} catch (Exception e) {
			e.printStackTrace();
			return new int[]{};
		}
	}
	
	protected static int[] getRevisionIDsTo(String articleTitle, int max, int to) {
		try {
			String url = "http://en.wikipedia.org/w/api.php?action=query&prop=revisions&titles=" + articleTitle + "&rvprop=ids&format=xml&rvlimit=" + max;
			Document doc = docBuilder.parse(url);
			NodeList revList = doc.getElementsByTagName("rev");
			int[] revs = new int[revList.getLength()];
			for (int i=0; i<revList.getLength(); i++) {
				String rev = revList.item(i).getAttributes().getNamedItem("revid").getNodeValue();
				revs[i] = Integer.parseInt(rev);
			}
			return revs;
		} catch (Exception e) {
			e.printStackTrace();
			return new int[]{};
		}
	}
	
	protected static int[] getRevisionIDsFromTo(String articleTitle, int from, int to) {
		try {
			String url = "http://en.wikipedia.org/w/api.php?action=query&prop=revisions&titles=" + articleTitle + "&rvprop=ids&format=xml";
			Document doc = docBuilder.parse(url);
			NodeList revList = doc.getElementsByTagName("rev");
			int[] revs = new int[revList.getLength()];
			for (int i=0; i<revList.getLength(); i++) {
				String rev = revList.item(i).getAttributes().getNamedItem("revid").getNodeValue();
				revs[i] = Integer.parseInt(rev);
			}
			return revs;
		} catch (Exception e) {
			e.printStackTrace();
			return new int[]{};
		}
	}
	

	protected static String getPlainTextArticleFromTitle(String title) {
		String url = "http://en.wikipedia.org/w/index.php?title=" + title + "&action=render";
		return getPlainTextArticleFromURL(url);
	}

	protected static String getPlainTextArticleFromID(int articleId) {
		String url = "http://en.wikipedia.org/w/index.php?oldid=" + articleId + "&action=render";
		return getPlainTextArticleFromURL(url);
	}

	/**
	 * 
	 */
	private static String getPlainTextArticleFromURL(String url) {
		try {
			if (INCLUDE_LINKS_IN_TEXT)
				url = "http://cgi.w3.org/cgi-bin/html2txt?url=" + URLEncoder.encode(url, "UTF-8") + "&noinlinerefs=on&endrefs=on";		
			String text;
			// -- Clean up
			if (INCLUDE_LINKS_IN_TEXT) {
				// The html is already plain text now
				text = getContent(url);
			} else {
				
				String html = getContent(url);
				
				html = removeElement(html, "table", null, "infobox");
				html = removeElement(html, "table", null, "navbox");
				html = removeElement(html, "table", "toc", null);

				// Remove [edit] sections in the html (also removes the extra space)
				html = html.replaceAll("<span class=\\\"editsection\\\">.*?<\\/span> ", "");
				
				// Convert
				text = HTMLUtil.textFromHTML(html);
				// Clean up line breaks in the text
				for (int i=0; i<10; i++)
					text = text.replace("\n  \n", "\n");
				for (int i=0; i<10; i++)
					text = text.replace("\n \n", "\n");
				for (int i=0; i<10; i++)
					text = text.replace("\n\n", "\n");
				text = text.replace("\n", "\n\n");
			}
			return text;
		} catch (Exception e) {
			e.printStackTrace();
			return e.getMessage();
		}
	}

	private static String getContent(String url) {
		try {
			System.out.print("\t.");
			URLConnection yc = (new URL(url)).openConnection();
			BufferedReader in = new BufferedReader(new InputStreamReader(yc.getInputStream()));
			String html = "";
			String inputLine;
			int nlines = 0;
			while ((inputLine = in.readLine()) != null) {
				nlines ++;
				if (nlines%50 == 0)
					System.out.print(".");
				html += inputLine + "\n";
			}
			in.close();
			System.out.println();
			return html;
		} catch (Exception e) {
			e.printStackTrace();
			return "";
		}
	}

	/**
	 * This ugly method is there to clean up HTML without having to parse the XML. XML parsers
	 * need well-formed HTML and wikipedia does not provide this. 
	 */
    public static String removeElement(String html, String element, String id, String cclass) {
    	String open = "<" + element;
    	if (id != null)
    		open += " id=\"" + id + "\"";
    	if (cclass != null)
    		open += " class=\"" + cclass;// + "\"";
    	String close = "</" + element + ">";
   	
        while (html.indexOf(open) != -1) {
        	int o_start = html.indexOf(open);
        	int o_end = 0;
        	
        	int tab = 1;
        	int from = o_start + open.length();
        	while (tab != 0) {
	        	int o = bestSearch(html.indexOf("<" + element + " ", from), html.indexOf("<" + element + ">", from));
	        	int c = html.indexOf(close, from);
	        	int best = bestSearch(o, c);
	        	if (c == -1)
	        		return null;
	        	if (best == o) {
	        		tab++;
	        		from = o + element.length() + 2;
	        	} else if (best == c) {
	        		tab--;
	        		from = c + element.length() + 2;
	        	} else
	        		return null;
	        	if (tab == 0) {
	        		o_end = c + close.length();
	        	}
        	}
//System.err.println("Removed " + element);
        			
        	html = html.substring(0, o_start) + html.substring(o_end);
        }
        
        return html;
    }
    
    private static int bestSearch(int i1, int i2) {
    	if (i1 == -1)
    		return i2;
    	if (i2 == -1)
    		return i1;
    	return Math.min(i1, i2);
    }
    
	private static String getRandomArticleTitle() {
		String articleURL = null;
		try {
			HttpURLConnection yc = (HttpURLConnection)(new URL("http://en.wikipedia.org/wiki/Special:Random")).openConnection();
			yc.getResponseCode(); // need to call this for redirection to work!
			articleURL = yc.getURL().toString(); // need to call this for redirection to work!
			String prefix = "http://en.wikipedia.org/wiki/";
			if (articleURL.startsWith(prefix))
				articleURL = articleURL.substring(prefix.length());
			else
				articleURL = null;
		} catch (Exception e) {
		}
		return articleURL;
	}
}
